from datetime import datetime

from oct_converter.dicom.metadata import (
    DicomMetadata,
    ImageGeometry,
    ManufacturerMeta,
    OCTDetectorType,
    OCTImageParams,
    OPTAcquisitionDevice,
    OPTAnatomyStructure,
    PatientMeta,
    SeriesMeta,
)
from oct_converter.image_types import OCTVolumeWithMetaData


def fds_patient_meta(fds_metadata: dict) -> PatientMeta:
    """Creates PatientMeta from FDS metadata

    Args:
        fds_metadata: Nested dictionary of metadata collected by the fds reader
    Returns:
        PatientMeta: Patient metadata populated by fds_metadata
    """
    patient_info = fds_metadata.get("patient_info_02") or fds_metadata.get(
        "patient_info", {}
    )
    sex_map = {1: "M", 2: "F", 3: "O", None: ""}
    patient = PatientMeta()

    patient.first_name = patient_info.get("first_name")
    patient.last_name = patient_info.get("last_name")
    patient.patient_id = patient_info.get("patient_id")
    patient.patient_sex = sex_map[patient_info.get("sex", None)]
    try:
        patient.patient_dob = datetime(*patient_info.get("birth_date"))
    except (TypeError, ValueError):
        pass

    return patient


def fds_series_meta(fds_metadata: dict) -> SeriesMeta:
    """Creates SeriesMeta from FDS metadata

    Args:
        fds_metadata: Nested dictionary of metadata collected by the fds reader
    Returns:
        SeriesMeta: Series metadata populated by fds_metadata
    """
    capture_info = fds_metadata.get("capture_info_02") or fds_metadata.get(
        "capture_info", {}
    )
    lat_map = {0: "R", 1: "L", None: ""}
    series = SeriesMeta()

    series.study_id = ""
    series.series_id = capture_info.get("session_id", "")
    series.laterality = lat_map[capture_info.get("eye", None)]
    series.acquisition_date = datetime(*capture_info.get("cap_date"))
    series.opt_anatomy = OPTAnatomyStructure.Retina

    return series


def fds_manu_meta(fds_metadata: dict, fds_header: dict) -> ManufacturerMeta:
    """Creates ManufacturerMeta from FDS metadata

    Args:
        fds_metadata: Nested dictionary of metadata collected by the fds reader
        fds_header: Dictionary of file header information parsed by the fda reader
    Returns:
        ManufacturerMeta: Manufacture metadata populated by fds_metadata
    """
    hw_info = (
        fds_metadata.get("hw_info_03")
        or fds_metadata.get("hw_info_02")
        or fds_metadata.get("hw_info_01", {})
    )
    manufacture = ManufacturerMeta()

    manufacture.manufacturer = "Topcon"
    manufacture.manufacturer_model = hw_info.get("model_name")
    manufacture.device_serial = hw_info.get("serial_number")
    manufacture.software_version = (
        f"{fds_header.get('major_ver')}.{fds_header.get('minor_ver')}"
    )

    return manufacture


def fds_image_geom(pixel_spacing: list) -> ImageGeometry:
    """Creates ImageGeometry from FDS metadata

    Args:
        pixel_spacing: Pixel spacing calculated in the fds reader
    Returns:
        ImageGeometry: Geometry data populated by pixel_spacing
    """
    image_geom = ImageGeometry()
    image_geom.pixel_spacing = [pixel_spacing[2], pixel_spacing[0]]
    image_geom.slice_thickness = pixel_spacing[1]
    image_geom.image_orientation = [1, 0, 0, 0, 1, 0]

    return image_geom


def fds_image_params() -> OCTImageParams:
    """Creates OCTImageParams specific to Topcon

    Args:
        None
    Returns:
        OCTImageParams: Image params populated with Topcon defaults
    """
    image_params = OCTImageParams()
    image_params.opt_acquisition_device = OPTAcquisitionDevice.OCTScanner
    image_params.DetectorType = OCTDetectorType.CCD
    image_params.IlluminationWaveLength = 830
    image_params.IlluminationPower = 800
    image_params.IlluminationBandwidth = 50
    image_params.DepthSpatialResolution = 6
    image_params.MaximumDepthDistortion = 0.5
    image_params.AlongscanSpatialResolution = 20
    image_params.MaximumAlongscanDistortion = 0.5
    image_params.AcrossscanSpatialResolution = 20
    image_params.MaximumAcrossscanDistortion = 0.5

    return image_params


def fds_dicom_metadata(oct: OCTVolumeWithMetaData) -> DicomMetadata:
    """Creates DicomMetadata and populates each module

    Args:
        oct: OCTVolumeWithMetaData created by the fds reader
    Returns:
        DicomMetadata: Populated DicomMetadata created with OCT metadata
    """
    meta = DicomMetadata
    meta.patient_info = fds_patient_meta(oct.metadata)
    meta.series_info = fds_series_meta(oct.metadata)
    meta.manufacturer_info = fds_manu_meta(oct.metadata, oct.header)
    meta.image_geometry = fds_image_geom(oct.pixel_spacing)
    meta.oct_image_params = fds_image_params()

    return meta



from __future__ import annotations

import warnings
from collections import defaultdict
from datetime import date, datetime, timedelta
from itertools import chain
from pathlib import Path

import numpy as np

from oct_converter.image_types import FundusImageWithMetaData, OCTVolumeWithMetaData
from oct_converter.readers.binary_structs import e2e_binary


class E2E(object):
    """Class for extracting data from Heidelberg's .e2e file format.

    Notes:
        Mostly based on description of .e2e file format here:
        https://bitbucket.org/uocte/uocte/wiki/Heidelberg%20File%20Format.

    Attributes:
        filepath: path to .e2e file for reading.
    """

    def __init__(self, filepath: str | Path) -> None:
        self.filepath = Path(filepath)
        if not self.filepath.exists():
            raise FileNotFoundError(self.filepath)

        self.power = pow(2, 10)
        self.sex = None
        self.first_name = None
        self.surname = None
        self.acquisition_date = None
        self.birthdate = None
        self.pixel_spacing = None

        # get initial directory structure
        with open(self.filepath, "rb") as f:
            raw = f.read(21)
            if raw == b"E2EMultipleVolumeFile":
                self.byte_skip = 64
            else:
                self.byte_skip = 0
            f.seek(self.byte_skip)
            raw = f.read(36)

            header = e2e_binary.header_structure.parse(raw)
            raw = f.read(52)
            main_directory = e2e_binary.main_directory_structure.parse(raw)

            # traverse list of main directories in first pass
            self.directory_stack = []

            current = main_directory.current
            while current != 0:
                self.directory_stack.append(current)
                f.seek(current + self.byte_skip)
                raw = f.read(52)
                directory_chunk = e2e_binary.main_directory_structure.parse(raw)
                current = directory_chunk.prev

    def read_oct_volume(
        self, legacy_intensity_transform: bool = False
    ) -> list[OCTVolumeWithMetaData]:
        """Reads OCT data.

        Args:
            legacy_intensity_transform: if True, use intensity transform used in v<=0.5.7. Defaults to False.

        Returns:
            A list of OCTVolumeWithMetaData.
        """

        def _make_lut():
            LUT = []
            for i in range(0, pow(2, 16)):
                LUT.append(self.uint16_to_ufloat16(i))
            return np.array(LUT)

        LUT = _make_lut()

        with open(self.filepath, "rb") as f:
            # get all subdirectories
            chunk_stack = []
            volume_dict = {}
            for position in self.directory_stack:
                f.seek(position + self.byte_skip)
                raw = f.read(52)
                directory_chunk = e2e_binary.main_directory_structure.parse(raw)

                for ii in range(directory_chunk.num_entries):
                    raw = f.read(44)
                    chunk = e2e_binary.sub_directory_structure.parse(raw)
                    volume_string = "{}_{}_{}".format(
                        chunk.patient_db_id, chunk.study_id, chunk.series_id
                    )
                    if volume_string not in volume_dict.keys():
                        volume_dict[volume_string] = chunk.slice_id / 2
                    elif chunk.slice_id / 2 > volume_dict[volume_string]:
                        volume_dict[volume_string] = chunk.slice_id / 2

                    if chunk.start > chunk.pos:
                        chunk_stack.append([chunk.start, chunk.size])

            # initalise dict to hold all the image volumes
            volume_array_dict = {}
            volume_array_dict_additional = (
                {}
            )  # for storage of slices not caught by extraction
            laterality_dict = {}
            laterality = None
            for volume, num_slices in volume_dict.items():
                if num_slices > 0:
                    # num_slices + 1 here due to evidence that a slice was being missed off the end in extraction
                    volume_array_dict[volume] = [0] * int(num_slices + 1)

            contour_dict = defaultdict(lambda: defaultdict(dict))

            # traverse all chunks and extract slices
            for start, pos in chunk_stack:
                f.seek(start + self.byte_skip)
                raw = f.read(60)
                chunk = e2e_binary.chunk_structure.parse(raw)

                if chunk.type == 9:  # patient data
                    raw = f.read(127)
                    try:
                        patient_data = e2e_binary.patient_id_structure.parse(raw)
                        self.sex = patient_data.sex
                        self.first_name = patient_data.first_name
                        self.surname = patient_data.surname
                        julian_birthdate = (patient_data.birthdate / 64) - 14558805
                        self.birthdate = self.julian_to_ymd(julian_birthdate)
                        # TODO: There are conflicting ideas of how to parse E2E's birthdate
                        # https://bitbucket.org/uocte/uocte/wiki/Heidelberg%20File%20Format suggests the above,
                        # whereas https://github.com/neurodial/LibE2E/blob/master/E2E/dataelements/patientdataelement.cpp
                        # suggests that DOB is given as a Windows date. Neither option seems accurate to
                        # test files with known-correct birthdates. More investigation is needed.
                        self.patient_id = patient_data.patient_id
                    except Exception:
                        pass

                elif chunk.type == 10004:  # bscan metadata
                    raw = f.read(104)
                    bscan_metadata = e2e_binary.bscan_metadata.parse(raw)
                    start_epoch = datetime(
                        year=1600, month=12, day=31, hour=23, minute=59
                    )
                    acquisition_datetime = start_epoch + timedelta(
                        seconds=bscan_metadata.acquisitionTime * 1e-7
                    )
                    if self.acquisition_date is None:
                        self.acquisition_date = acquisition_datetime.date()
                    if self.pixel_spacing is None:
                        # scaley found, x and z not yet found in file
                        # but taken from E2E reader settings
                        self.pixel_spacing = [0.011484, bscan_metadata.scaley, 0.244673]

                elif chunk.type == 11:  # laterality data
                    raw = f.read(20)
                    try:
                        laterality_data = e2e_binary.lat_structure.parse(raw)
                        if laterality_data.laterality == 82:
                            laterality = "R"
                        elif laterality_data.laterality == 76:
                            laterality = "L"
                    except Exception:
                        laterality = None

                elif chunk.type == 10019:  # contour data
                    raw = f.read(16)
                    contour_data = e2e_binary.contour_structure.parse(raw)

                    if contour_data.width > 0:
                        volume_string = "{}_{}_{}".format(
                            chunk.patient_db_id, chunk.study_id, chunk.series_id
                        )
                        slice_id = int(chunk.slice_id / 2) - 1
                        contour_name = f"contour{contour_data.id}"
                        try:
                            raw_volume = np.frombuffer(
                                f.read(contour_data.width * 4), dtype=np.float32
                            )
                            contour = np.array(raw_volume)
                            max_float = np.finfo(np.float32).max
                            contour[(contour < 1e-9) | (contour == max_float)] = np.nan
                        except Exception as e:
                            warnings.warn(
                                (
                                    f"Could not read contour "
                                    f"image id {volume_string}"
                                    f"contour name {contour_name} "
                                    f"slice id {slice_id}."
                                ),
                                UserWarning,
                            )
                        else:
                            (
                                contour_dict[volume_string][contour_name][slice_id]
                            ) = contour

                elif chunk.type == 1073741824:  # image data
                    raw = f.read(20)
                    image_data = e2e_binary.image_structure.parse(raw)

                    if chunk.ind == 1:  # oct data
                        count = image_data.height * image_data.width
                        if count == 0:
                            break
                        raw_volume = np.fromfile(f, dtype=np.uint16, count=count)
                        volume_string = "{}_{}_{}".format(
                            chunk.patient_db_id, chunk.study_id, chunk.series_id
                        )
                        try:
                            image = LUT[raw_volume].reshape(
                                image_data.height, image_data.width
                            )
                        except Exception:
                            warnings.warn(
                                (
                                    f"Could not reshape image id {volume_string} with "
                                    f"{len(LUT[raw_volume])} elements into a "
                                    f"{image_data.height}x"
                                    f"{image_data.width} array"
                                ),
                                UserWarning,
                            )
                        else:
                            if legacy_intensity_transform:
                                image = pow(image, 1.0 / 2.4)
                            else:
                                image = self.vol_intensity_transform(image)

                            if volume_string in volume_array_dict.keys():
                                volume_array_dict[volume_string][
                                    int(chunk.slice_id / 2) - 1
                                ] = image
                            else:
                                # try to capture these additional images
                                if volume_string in volume_array_dict_additional.keys():
                                    volume_array_dict_additional[volume_string].append(
                                        image
                                    )
                                else:
                                    volume_array_dict_additional[volume_string] = [
                                        image
                                    ]
                            # here assumes laterality stored in chunk before the image itself
                            if laterality and volume_string not in laterality_dict:
                                laterality_dict[volume_string] = laterality

            contour_data = {}
            for volume_id, contours in contour_dict.items():
                if volume_id in volume_dict:
                    num_slices = int(volume_dict[volume_id]) + 1
                else:
                    num_slices = None
                contour_data[volume_id] = {
                    k: [None] * (num_slices or len(v)) for k, v in contours.items()
                }

                for contour_name, contour_values in contours.items():
                    for slice_id, contour in contour_values.items():
                        (contour_data[volume_id][contour_name][slice_id]) = contour

            # Read metadata to attach to OCTVolumeWithMetaData
            metadata = self.read_all_metadata()

            oct_volumes = []
            for key, volume in chain(
                volume_array_dict.items(), volume_array_dict_additional.items()
            ):
                # remove any initalised volumes that never had image data attached
                volume = [slc for slc in volume if not isinstance(slc, int)]
                if volume is None or len(volume) == 0:
                    continue
                oct_volumes.append(
                    OCTVolumeWithMetaData(
                        volume=volume,
                        patient_id=self.patient_id,
                        first_name=self.first_name,
                        surname=self.surname,
                        sex=self.sex,
                        patient_dob=self.birthdate,
                        acquisition_date=self.acquisition_date,
                        volume_id=key,
                        laterality=laterality_dict.get(key),
                        contours=contour_data.get(key),
                        pixel_spacing=self.pixel_spacing,
                        metadata=metadata,
                    )
                )

        return oct_volumes

    def read_fundus_image(self) -> list[FundusImageWithMetaData]:
        """Reads fundus data.

        Returns:
            A sequence of FundusImageWithMetaData.
        """
        with open(self.filepath, "rb") as f:
            # traverse in second pass and  get all subdirectories
            chunk_stack = []
            for position in self.directory_stack:
                f.seek(position + self.byte_skip)
                raw = f.read(52)
                directory_chunk = e2e_binary.main_directory_structure.parse(raw)

                for ii in range(directory_chunk.num_entries):
                    raw = f.read(44)
                    chunk = e2e_binary.sub_directory_structure.parse(raw)
                    if chunk.start > chunk.pos:
                        chunk_stack.append([chunk.start, chunk.size])

            # initalise dict to hold all the image volumes
            image_array_dict = {}
            laterality_dict = {}

            # traverse all chunks and extract slices
            for start, pos in chunk_stack:
                f.seek(start + self.byte_skip)
                raw = f.read(60)
                chunk = e2e_binary.chunk_structure.parse(raw)

                if chunk.type == 9:  # patient data
                    raw = f.read(127)
                    try:
                        patient_data = e2e_binary.patient_id_structure.parse(raw)
                        self.sex = patient_data.sex
                        self.first_name = patient_data.first_name
                        self.surname = patient_data.surname
                        julian_birthdate = (patient_data.birthdate / 64) - 14558805
                        self.birthdate = self.julian_to_ymd(julian_birthdate)
                        self.patient_id = patient_data.patient_id
                    except Exception:
                        pass

                if chunk.type == 11:  # laterality data
                    raw = f.read(20)
                    try:
                        laterality_data = e2e_binary.lat_structure.parse(raw)
                        if laterality_data.laterality == 82:
                            laterality = "R"
                        elif laterality_data.laterality == 76:
                            laterality = "L"
                    except Exception:
                        laterality = None

                if chunk.type == 1073741824:  # image data
                    raw = f.read(20)
                    image_data = e2e_binary.image_structure.parse(raw)
                    count = image_data.height * image_data.width
                    if count == 0:
                        break
                    if chunk.ind == 0:  # fundus data
                        raw_volume = np.frombuffer(f.read(count), dtype=np.uint8)
                        image = np.array(raw_volume).reshape(
                            image_data.height, image_data.width
                        )
                        image_string = "{}_{}_{}".format(
                            chunk.patient_db_id, chunk.study_id, chunk.series_id
                        )
                        image_array_dict[image_string] = image
                        # here assumes laterality stored in chunk before the image itself
                        laterality_dict[image_string] = laterality

            # Read metadata to attach to FundusImageWithMetaData
            metadata = self.read_all_metadata()

            fundus_images = []
            for key, image in image_array_dict.items():
                fundus_images.append(
                    FundusImageWithMetaData(
                        image=image,
                        patient_id=self.patient_id,
                        image_id=key,
                        laterality=laterality_dict[key]
                        if key in laterality_dict.keys()
                        else None,
                        metadata=metadata,
                    )
                )

        return fundus_images

    def read_all_metadata(self):
        """
        Reads all available metadata and returns a dictionary.

        The metadata is a raw dump of everything available.

        Returns:
            dictionary with all metadata.
        """

        def _convert_to_dict(container):
            """Converts a container object to a dictionary"""
            return dict(
                (name, getattr(container, name))
                for name in container
                if not name.startswith("_")
            )

        metadata = dict()
        metadata["image_data"] = []
        metadata["bscan_data"] = []
        metadata["patient_data"] = []
        metadata["laterality_data"] = []
        metadata["contour_data"] = []
        metadata["fundus_data"] = []

        with open(self.filepath, "rb") as f:
            # get all subdirectories
            chunk_stack = []

            for position in self.directory_stack:
                f.seek(position + self.byte_skip)
                raw = f.read(52)
                directory_chunk = e2e_binary.main_directory_structure.parse(raw)

                for ii in range(directory_chunk.num_entries):
                    raw = f.read(44)
                    chunk = e2e_binary.sub_directory_structure.parse(raw)
                    if chunk.start > chunk.pos:
                        chunk_stack.append([chunk.start, chunk.size])

            # traverse all chunks and extract slices
            for start, pos in chunk_stack:
                f.seek(start + self.byte_skip)
                raw = f.read(60)
                chunk = e2e_binary.chunk_structure.parse(raw)

                if chunk.type == 9:  # patient data
                    raw = f.read(127)
                    try:
                        patient_data = e2e_binary.patient_id_structure.parse(raw)
                        metadata["patient_data"].append(_convert_to_dict(patient_data))
                    except Exception:
                        pass

                elif chunk.type == 10004:  # bscan metadata
                    raw = f.read(104)
                    bscan_metadata = e2e_binary.bscan_metadata.parse(raw)
                    metadata["bscan_data"].append(_convert_to_dict(bscan_metadata))

                if chunk.type == 1073741824:  # fundus data
                    raw = f.read(20)
                    fundus_data = e2e_binary.image_structure.parse(raw)
                    metadata["fundus_data"].append(_convert_to_dict(fundus_data))

                elif chunk.type == 11:  # laterality data
                    raw = f.read(20)
                    laterality_data = e2e_binary.lat_structure.parse(raw)
                    metadata["laterality_data"].append(
                        _convert_to_dict(laterality_data)
                    )

                elif chunk.type == 10019:  # contour data
                    raw = f.read(16)
                    contour_data = e2e_binary.contour_structure.parse(raw)
                    metadata["contour_data"].append(_convert_to_dict(contour_data))

                elif chunk.type == 1073741824:  # image data
                    raw = f.read(20)
                    image_data = e2e_binary.image_structure.parse(raw)
                    metadata["image_data"].append(_convert_to_dict(image_data))

        return metadata

    def read_custom_float(self, bytes: str) -> float:
        """Implementation of bespoke float type used in .e2e files.

        Notes:
            Custom float is a floating point type with no sign, 6-bit exponent, and 10-bit mantissa.

        Args:
            bytes: the two bytes.

        Returns:
            float
        """
        # convert two bytes to 16-bit binary representation
        bits = bin(bytes[0])[2:].zfill(8)[::-1] + bin(bytes[1])[2:].zfill(8)[::-1]

        # get mantissa and exponent
        mantissa = bits[:10]
        exponent = bits[10:]

        # convert to decimal representations
        mantissa_sum = 1 + int(mantissa, 2) / self.power
        exponent_sum = int(exponent[::-1], 2) - 63
        decimal_value = mantissa_sum * pow(2, exponent_sum)
        return decimal_value

    def uint16_to_ufloat16(self, uint16: int) -> float:
        """Implementation of bespoke float type used in .e2e files.

        Notes:
            Custom float is a floating point type with no sign, 6-bit exponent, and 10-bit mantissa.

        Args:
            uint16

        Returns:
            float
        """
        bits = "{0:016b}".format(uint16)[::-1]
        # get mantissa and exponent
        mantissa = bits[:10]
        exponent = bits[10:]
        exponent = exponent[::-1]

        # convert to decimal representations
        mantissa_sum = 1 + int(mantissa, 2) / self.power
        exponent_sum = int(exponent, 2) - 63
        decimal_value = mantissa_sum * np.float_power(2, exponent_sum)
        return decimal_value

    def vol_intensity_transform(self, data: np.array) -> np.array:
        """Implementation of intensity transform used in .e2e files.

        Notes:
            Code thanks to @oli4, see discussion in https://github.com/marksgraham/OCT-Converter/issues/21#issuecomment-1057455183
        """
        selection_0 = data == np.finfo(np.float32).max
        selection_data = data <= 1

        new = np.log(data[selection_data] + 2.44e-04)
        new = (new + 8.3) / 8.285

        data[selection_data] = new
        data[selection_0] = 0
        data = np.clip(data, 0, 1)
        return data

    def julian_to_ymd(self, J):
        """Converts Julian Day to Gregorian YMD.

        see https://en.wikipedia.org/wiki/Julian_day
        with thanks to https://github.com/seanredmond/juliandate
        """
        y = 4716
        j = 1401
        m = 2
        n = 12
        r = 4
        p = 1461
        v = 3
        u = 5
        s = 153
        w = 2
        B = 274277
        C = -38

        f = J + j + int(((int((4 * J + B) / 146097)) * 3) / 4) + C
        e = r * f + v
        g = int((e % p) / r)
        h = u * g + w

        D = int((h % s) / u) + 1
        M = ((int(h / s) + m) % n) + 1
        Y = int(e / p) - y + int((n + m - M) / n)

        return date(Y, M, D)



import json

from oct_converter.dicom import create_dicom_from_oct
from oct_converter.readers import FDS

# An example .fds file can be downloaded from the Biobank website:
# https://biobank.ndph.ox.ac.uk/showcase/refer.cgi?id=30
filepath = "/home/mark/Downloads/eg_oct_fds.fds"
fds = FDS(filepath)

oct_volume = (
    fds.read_oct_volume()
)  # returns an OCT volume with additional metadata if available
oct_volume.peek()  # plots a montage of the volume
oct_volume.save("fds_testing.avi")  # save volume as a movie
oct_volume.save(
    "fds_testing.png"
)  # save volume as a set of sequential images, fds_testing_[1...N].png
oct_volume.save_projection("projection.png")

fundus_image = (
    fds.read_fundus_image()
)  # returns a  Fundus image with additional metadata if available
fundus_image.save("fds_testing_fundus.jpg")

# extract all other metadata
metadata = fds.read_all_metadata(verbose=True)
with open("fds_metadata.json", "w") as outfile:
    outfile.write(json.dumps(metadata, indent=4))

# create a DICOM from FDS
dcm = create_dicom_from_oct(filepath)
# Output dir can be specified, otherwise will
# default to current working directory.



from __future__ import annotations

import re
from datetime import datetime
from pathlib import Path

import numpy as np

from oct_converter.image_types import OCTVolumeWithMetaData


class POCT(object):
    """Class for extracting data from Optovues's .oct file format.

    Attributes:
        filepath: path to .oct file for reading. Expects a file with the same name and a .txt extension
            exists at the same location.
    """

    def __init__(self, filepath: str | Path) -> None:
        self.filepath = Path(filepath)
        if not self.filepath.exists():
            raise FileNotFoundError(self.filepath)
        self.filespec = self.filepath.with_suffix(".txt")
        if not self.filespec.exists():
            raise FileNotFoundError(
                f"Could not find filespec {self.filespec} in the same location as {self.filepath}"
            )

    def _read_filespec(self) -> None:
        scan_info = []
        file_info = {}
        with open(self.filespec, "r", encoding="iso-8859-1") as f:
            lines = f.readlines()
            for i, line in enumerate(lines):
                if "Window Height" in line:
                    height = [int(s) for s in line.split() if s.isdigit()][0]
                if "Scan Length" in line and "Scan Usage" in lines[i + 1]:
                    scan_length = [int(s) for s in line.split() if s.isdigit()][0]
                    scan_number = [int(s) for s in lines[i + 1].split() if s.isdigit()][
                        0
                    ]
                    scan_info.append(
                        {"height": height, "length": scan_length, "number": scan_number}
                    )
                if "Eye Scanned" in line:
                    if "OD" in line:
                        file_info["laterality"] = "R"
                    elif "OS" in line:
                        file_info["laterality"] = "L"
                if "Video Height" in line:
                    file_info["video_height"] = line.split("=")[-1]
                if "Video Width" in line:
                    file_info["video_width"] = line.split("=")[-1]
                if "BitCount" in line:
                    file_info["bitcount"] = line.split("=")[-1]
                if "Physical video width" in line:
                    file_info["physical_width"] = line.split("=")[-1]
                if "Physical video Height" in line:
                    file_info["physical_height"] = line.split("=")[-1]

        if "video_height" in file_info and "physical_height" in file_info:
            file_info["scale_y"] = float(file_info["physical_height"].split()[0]) / abs(
                float(file_info["video_height"])
            )
        if "video_width" in file_info and "physical_width" in file_info:
            file_info["scale_x"] = float(file_info["physical_width"].split()[0]) / abs(
                float(file_info["video_width"])
            )

        # Attempt to find acquisition date in filename
        acq = (
            list(
                re.search(
                    r"(?P<y>\d{4})-(?P<m>\d{1,2})-(?P<d>\d{1,2})_(?P<h>\d{1,2}).(?P<M>\d{1,2}).(?P<s>\d{1,2})",
                    str(self.filespec),
                ).groups()
            )
            if re.search(
                r"(?P<y>\d{4})-(?P<m>\d{1,2})-(?P<d>\d{1,2})_(?P<h>\d{1,2}).(?P<M>\d{1,2}).(?P<s>\d{1,2})",
                str(self.filespec),
            )
            else None
        )
        if acq:
            file_info["acquisition_date"] = datetime(
                year=int(acq[0]),
                month=int(acq[1]),
                day=int(acq[2]),
                hour=int(acq[3]),
                minute=int(acq[4]),
                second=int(acq[5]),
            )

        self.scan_info = scan_info
        self.file_info = file_info

    def read_oct_volume(self) -> list[OCTVolumeWithMetaData]:
        """Reads OCT data.

        Returns:
            OCTVolumeWithMetaData
        """
        self._read_filespec()

        with open(self.filepath, "rb") as f:
            data = np.frombuffer(
                f.read(), dtype=np.float32
            )  # np.fromstring() gives numpy depreciation warning
            all_volumes = []
            for volume in self.scan_info:
                num_pixels_slice = volume["height"] * volume["length"]
                num_slices = volume["number"]
                all_slices = []
                for i in range(num_slices):
                    slice = data[i * num_pixels_slice : (i + 1) * num_pixels_slice]
                    slice = np.rot90(slice.reshape(volume["length"], volume["height"]))
                    all_slices.append(slice)
                all_volumes.append(
                    OCTVolumeWithMetaData(
                        all_slices,
                        acquisition_date=self.file_info.get("acquisition_date", None),
                        laterality=self.file_info.get("laterality", ""),
                        pixel_spacing=[
                            self.file_info.get("scale_x", 0.015),
                            self.file_info.get("scale_y", 0.015),
                        ],
                    )
                )
        return all_volumes



from oct_converter.dicom import create_dicom_from_oct
from oct_converter.readers import IMG

filepath = "../sample_files/file.img"
img = IMG(filepath)
oct_volume = (
    img.read_oct_volume()
)  # returns an OCT volume with additional metadata if available
oct_volume.peek()  # plots a montage of the volume
oct_volume.save("img_testing.avi")  # save volume

# create a DICOM from .img
dcm = create_dicom_from_oct(filepath)
# Output dir can be specified, otherwise will
# default to current working directory.
# Additionally, rows, columns, and interlaced can
# be specified to more accurately create an image.
dcm = create_dicom_from_oct(filepath, interlaced=True)



from __future__ import annotations

from datetime import datetime

from oct_converter.dicom.metadata import (
    DicomMetadata,
    ImageGeometry,
    ManufacturerMeta,
    OCTDetectorType,
    OCTImageParams,
    OPTAcquisitionDevice,
    OPTAnatomyStructure,
    PatientMeta,
    SeriesMeta,
)
from oct_converter.image_types import OCTVolumeWithMetaData


def img_patient_meta(patient_id: str = "") -> PatientMeta:
    """Creates PatientMeta populated with id if known from filename

    Args:
        patient_id: Patient ID, parsed from filename
    Returns:
        PatientMeta: Empty (or mostly empty) PatientMeta
    """
    patient = PatientMeta()

    patient.first_name = ""
    patient.last_name = ""
    patient.patient_id = patient_id
    patient.patient_sex = ""
    patient.patient_dob = None

    return patient


def img_series_meta(
    acquisition_date: datetime | None = None,
    laterality: str = "",
) -> SeriesMeta:
    """Creates SeriesMeta popualted with aquisition_date and
    laterality if known

    Args:
        acquisition_date: Date of acquisition, parsed from filename
        laterality: R or L, parsed from filename OD or OS
    Returns:
        SeriesMeta: Mostly empty SeriesMeta
    """
    series = SeriesMeta()

    series.study_id = ""
    series.series_id = 0
    series.laterality = laterality
    series.acquisition_date = acquisition_date
    series.opt_anatomy = OPTAnatomyStructure.Retina

    return series


def img_manu_meta() -> ManufacturerMeta:
    """Because img file contains no manufacture data,
    creates ManufacturerMeta with only manufacturer.

    Args:
        None
    Returns:
        ManufacturerMeta: Mostly empty ManufacturerMeta
    """
    manufacture = ManufacturerMeta()

    manufacture.manufacturer = "Zeiss"
    manufacture.manufacturer_model = ""
    manufacture.device_serial = ""
    manufacture.software_version = ""

    return manufacture


def img_image_geom() -> ImageGeometry:
    """Creates ImageGeometry with generic values

    Args:
        None
    Returns:
        ImageGeometry: Geometry data populated with generic values
    """
    image_geom = ImageGeometry()
    image_geom.pixel_spacing = [0.002, 0.002]
    image_geom.slice_thickness = 0.004
    image_geom.image_orientation = [1, 0, 0, 0, 1, 0]

    return image_geom


def img_image_params() -> OCTImageParams:
    """Creates OCTImageParams specific to Zeiss

    Args:
        None
    Returns:
        OCTImageParams: Image params populated with img defaults
    """
    image_params = OCTImageParams()
    image_params.opt_acquisition_device = OPTAcquisitionDevice.OCTScanner
    image_params.DetectorType = OCTDetectorType.CCD
    image_params.IlluminationWaveLength = 830
    image_params.IlluminationPower = 800
    image_params.IlluminationBandwidth = 50
    image_params.DepthSpatialResolution = 6
    image_params.MaximumDepthDistortion = 0.5
    image_params.AlongscanSpatialResolution = []
    image_params.MaximumAlongscanDistortion = []
    image_params.AcrossscanSpatialResolution = []
    image_params.MaximumAcrossscanDistortion = []

    return image_params


def img_dicom_metadata(oct: OCTVolumeWithMetaData) -> DicomMetadata:
    """Creates DicomMetadata and populates each module

    Args:
        oct: OCTVolumeWithMetaData created by img reader
    Returns:
        DicomMetadata: Populated DicomMetadata created with Zeiss defaults
        and information extracted from filename if able
    """
    meta = DicomMetadata
    meta.patient_info = img_patient_meta(oct.patient_id)
    meta.series_info = img_series_meta(oct.acquisition_date, oct.laterality)
    meta.manufacturer_info = img_manu_meta()
    meta.image_geometry = img_image_geom()
    meta.oct_image_params = img_image_params()

    return meta



from oct_converter.dicom.metadata import (
    DicomMetadata,
    ImageGeometry,
    ManufacturerMeta,
    OCTDetectorType,
    OCTImageParams,
    OPTAcquisitionDevice,
    OPTAnatomyStructure,
    PatientMeta,
    SeriesMeta,
)
from oct_converter.image_types import OCTVolumeWithMetaData


def boct_series_meta(boct: OCTVolumeWithMetaData) -> SeriesMeta:
    """Creates SeriesMeta from Bioptigen OCT metadata

    Args:
        boct: OCTVolumeWithMetaData with laterality and acquisition_date attributes
    Returns:
        SeriesMeta: Series metadata populated with laterality and acquisition_date
    """
    series = SeriesMeta()

    series.study_id = ""
    series.series_id = 0
    series.laterality = boct.laterality
    series.acquisition_date = boct.acquisition_date
    series.opt_anatomy = OPTAnatomyStructure.Retina

    return series


def boct_manu_meta() -> ManufacturerMeta:
    """Creates base ManufacturerMeta for Bioptigen

    Args:
        None
    Returns:
        ManufacturerMeta: base Bioptigen manufacture metadata
    """

    manufacture = ManufacturerMeta()

    manufacture.manufacturer = "Bioptigen"
    manufacture.manufacturer_model = ""
    manufacture.device_serial = ""
    manufacture.software_version = ""

    return manufacture


def boct_image_geom() -> ImageGeometry:
    """Creates ImageGeometry from Bioptigen OCT metadata

    Args:
        pixel_spacing: Pixel spacing calculated in the boct reader
    Returns:
        ImageGeometry: Geometry data populated by pixel_spacing
    """
    image_geom = ImageGeometry()
    image_geom.pixel_spacing = [0.02, 0.02]  # Placeholder value
    image_geom.slice_thickness = 0.2  # Placeholder value
    image_geom.image_orientation = [1, 0, 0, 0, 1, 0]

    return image_geom


def boct_image_params() -> OCTImageParams:
    """Creates OCTImageParams specific to Bioptigen

    Args:
        None
    Returns:
        OCTImageParams: Image params populated with Bioptigen defaults
    """
    image_params = OCTImageParams()
    image_params.opt_acquisition_device = OPTAcquisitionDevice.OCTScanner
    image_params.DetectorType = OCTDetectorType.CCD
    image_params.IlluminationWaveLength = 880
    image_params.IlluminationPower = 1200
    image_params.IlluminationBandwidth = 50
    image_params.DepthSpatialResolution = 7
    image_params.MaximumDepthDistortion = 0.5
    image_params.AlongscanSpatialResolution = []
    image_params.MaximumAlongscanDistortion = []
    image_params.AcrossscanSpatialResolution = []
    image_params.MaximumAcrossscanDistortion = []

    return image_params


def boct_dicom_metadata(boct: OCTVolumeWithMetaData) -> DicomMetadata:
    """Creates DicomMetadata and populates each module

    Args:
        oct: OCTVolumeWithMetaData created by the boct reader
    Returns:
        DicomMetadata: Populated DicomMetadata created with OCT metadata
    """
    meta = DicomMetadata
    meta.patient_info = PatientMeta()
    meta.series_info = boct_series_meta(boct)
    meta.manufacturer_info = boct_manu_meta()
    meta.image_geometry = boct_image_geom()
    meta.oct_image_params = boct_image_params()

    return meta



from oct_converter.dicom import create_dicom_from_oct
from oct_converter.readers import POCT

filepath = "../sample_files/sample.OCT"
poct = POCT(filepath)
oct_volumes = poct.read_oct_volume()

for volume in oct_volumes:
    volume.peek()  # plots a montage of the volume
print("debug")

# create DICOM from .OCT
dcm = create_dicom_from_oct(filepath)
# Output dir can be specified, otherwise will
# default to current working directory.
# If multiple volumes are identified within the file,
# multiple DICOMs will be outputted.



from construct import (
    Array,
    Float32n,
    Float64n,
    Int8un,
    Int16un,
    Int32un,
    PaddedString,
    Struct,
    this,
)

"""
        Notes:
        Mostly based on description of .fds file format here:
        https://bitbucket.org/uocte/uocte/wiki/Topcon%20File%20Format

        header (obj:Struct) : Defines structure of volume's header.
        oct_header (obj:Struct) : Defines structure of OCT header (IMG_SCAN_03).
        oct_header_2 (obj:Struct) : Defines structure of OCT header (IMG_SCAN_02).
        fundus_header (obj:Struct): Defines structure of fundus header (IMG_OBS).
        param_scan_04_header (obj:Struct) : Defines PARAM_SCAN_04 header.
        param_scan_02_header (obj:Struct) : Defines PARAM_SCAN_02 header.
        hw_info_03_header (obj:Struct) : Defines HW_INFO_03 header.
        hw_info_02_header (obj:Struct) : Defines HW_INFO_02 header.
        hw_info_01_header (obj:Struct) : Defines HW_INFO_01 header.
        patient_info_02_header (obj:Struct) : Defines PATIENT_INFO_02 header.
        file_info_header (obj:Struct) : Defines FILE_INFO header.
        capture_info_02_header (obj:Struct) : Defines CAPTURE_INFO_02 header.
        capture_info_header (obj:Struct) : Defines CAPTURE_INFO header.
        img_trc_02_header (obj:Struct) : Defines IMG_TRC_02 header.
        img_trc_header (obj:Struct) : Defines IMG_TRC header.
        param_obs_02_header (obj:Struct) : Defines PARAM_OBS_02 when size is 90.
        param_obs_02_short_header (obj:Struct) : Defines PARAM_OBS_02 when size is 6.
        param_trc_02_header (obj:Struct) : Defines PARAM_TRC_02 header.
        img_mot_comp_03_header (obj:Struct) : Defines IMG_MOT_COMP_03 header.
        img_mot_comp_02_header (obj:Struct) : Defines IMG_MOT_COMP_02 header.
        img_mot_comp_header (obj:Struct) : Defines IMG_MOT_COMP header.
        img_projection_header (obj:Struct) : Defines IMG_PROJECTION header.
        effective_scan_range_header (obj:Struct) : Defines EFFECTIVE_SCAN_RANGE header.
        regist_info_header (obj:Struct) : Defines REGIST_INFO header.
        regist_scan_02_header (obj:Struct) : Defines REGIST_SCAN_02 header.
        result_cornea_curve_header (obj:Struct) : Defines RESULT_CORNEA_CURVE header.
        result_cornea_thickness_header (obj:Struct) : Defines RESULT_CORNEA_THICKNESS header.
        align_info_header (obj:Struct) : Defines ALIGN_INFO header.
        main_module_info_header (obj:Struct) : Defines MAIN_MODULE_INFO header.
        fast_q2_info_header (obj:Struct) : Defines FAST_Q2_INFO header.
        gla_littmann_01_header (obj:Struct) : Defines GLA_LITTMANN_01 header.
        thumbnail_header (obj:Struct) : Defines THUMBNAIL header.
        patientext_info_header (obj:Struct) : Defines PATIENTEXT_INFO header.
"""

header = Struct(
    "file_code" / PaddedString(4, "ascii"),  # Always "FOCT"
    "file_type" / PaddedString(3, "ascii"),
    # file_type is "FDA" or "FAA",
    # denoting "macula" or "external" fixation
    "major_ver" / Int32un,
    "minor_ver" / Int32un,
)

# IMG_SCAN_03
oct_header = Struct(
    "scan_mode" / Int8un,  # 2 = 3D, 3 = Radial, 4 = Cross
    "width" / Int32un,
    "height" / Int32un,
    "bits_per_pixel" / Int32un,
    "number_slices" / Int32un,
    "format" / Int8un,
    "size" / Int32un,
)

# IMG_SCAN_02
oct_header_2 = Struct(
    "scan_mode" / Int8un,  # 2 = 3D, 3 = Radial, 4 = Cross
    "width" / Int32un,
    "height" / Int32un,
    "bits_per_pixel" / Int32un,
    "number_slices" / Int32un,
    "fast_q" / Float64n,
    "format" / Int8un,
    "size" / Int32un,
)

# IMG_OBS
fundus_header = Struct(
    "width" / Int32un,
    "height" / Int32un,
    "bits_per_pixel" / Int32un,
    "number_slices" / Int32un,
    "format" / Int8un,
    "size" / Int32un,
)

param_scan_04_header = Struct(
    "fixation" / Int32un,
    "mirror_pos" / Int32un,
    "polar" / Int32un,
    "x_dimension_mm" / Float64n,
    "y_dimension_mm" / Float64n,
    "z_resolution_um" / Float64n,
    "comp_eff_2" / Float64n,
    "comp_eff_3" / Float64n,
    "base_pos" / Int8un,
    "used_calib_data" / Int8un,
)

param_scan_02_header = Struct(
    "scan_mode" / Int8un,
    "light_level" / Int32un,
    "fixation" / Int32un,
    "mirror_pos" / Int32un,
    "nd" / Int32un,
    "polar" / Int32un,
    "x_dimension_mm" / Float64n,
    "y_dimension_mm" / Float64n,
    "z_resolution_um" / Float64n,
    "comp_eff_2" / Float64n,
    "comp_eff_3" / Float64n,
    "noise_thresh" / Float64n,
    "range_adj" / Float64n,
    "base_pos" / Int8un,
)

hw_info_03_header = Struct(
    "model_name" / PaddedString(16, "ascii"),
    "serial_number" / PaddedString(16, "ascii"),
    "spect_sn" / PaddedString(16, "ascii"),
    "rom_ver" / PaddedString(16, "ascii"),
    "unknown" / PaddedString(16, "ascii"),
    "eq_calib_year" / Int16un,
    "eq_calib_month" / Int16un,
    "eq_calib_day" / Int16un,
    "eq_calib_hour" / Int16un,
    "eq_calib_minute" / Int16un,
    "spect_calib_year" / Int16un,
    "spect_calib_month" / Int16un,
    "spect_calib_day" / Int16un,
    "spect_calib_hour" / Int16un,
    "spect_calib_minute" / Int16un,
)

hw_info_02_header = Struct(
    "model_name" / PaddedString(16, "ascii"),
    "serial_number" / PaddedString(16, "ascii"),
    "spect_sn" / PaddedString(16, "ascii"),
    "rom_ver" / PaddedString(16, "ascii"),
    "eq_calib_year" / Int16un,
    "eq_calib_month" / Int16un,
    "eq_calib_day" / Int16un,
    "spect_calib_year" / Int16un,
    "spect_calib_month" / Int16un,
    "spect_calib_day" / Int16un,
)

hw_info_01_header = Struct(
    "model_name" / PaddedString(16, "ascii"),
    "serial_number" / PaddedString(16, "ascii"),
    "spect_sn" / PaddedString(16, "ascii"),
    "rom_ver" / PaddedString(16, "ascii"),
    "eq_calib_year" / Int16un,
    "eq_calib_month" / Int16un,
    "eq_calib_day" / Int16un,
    "spect_calib_year" / Int16un,
    "spect_calib_month" / Int16un,
    "spect_calib_day" / Int16un,
)

patient_info_02_header = Struct(
    "patient_id" / PaddedString(32, "ascii"),
    "first_name" / PaddedString(32, "ascii"),
    "last_name" / PaddedString(32, "ascii"),
    "mid_name" / PaddedString(8, "ascii"),
    "sex" / Int8un,
    "birth_date" / Int16un[3],
    "occup_reg" / Int8un[20][2],
    "r_date" / Int16un[3],
    "record_id" / Int8un[16],
    "lv_date" / Int16un[3],
    # I've not found files that have the below information,
    # so it's difficult to confirm the remaining.
    "physician" / Int8un[64][2],  # [64 2] ???
    "zip_code" / Int8un[12],  # how does this make sense.
    "addr" / Int8un[48][2],  # [48 2]
    "phones" / Int8un[16][2],  # [16 2]
    "nx_date" / Int16un[6],  # [1 6]
    "multipurpose_field" / Int8un[20][3],  # [20 3]
    "descp" / Int8un[64],
    "reserved" / Int8un[32],
)

file_info_header = Struct(
    "0x2" / Int32un,
    "0x3e8" / Int32un,
    "8.0.1.20198" / PaddedString(32, "ascii"),
)

capture_info_02_header = Struct(
    "eye" / Int8un,
    "scan_mode" / Int8un,
    "session_id" / Int32un,
    "label" / PaddedString(100, "ascii"),
    "cap_date" / Int16un[6],
)

capture_info_header = Struct(
    "eye" / Int8un,
    "cap_date" / Int16un[6],
)

img_trc_02_header = Struct(
    "width" / Int32un,
    "height" / Int32un,
    "bits_per_pixel" / Int32un,
    "num_slices_0x2" / Int32un,
    "format" / Int8un,
    "size" / Int32un,
)

img_trc_header = Struct(
    "width" / Int32un,
    "height" / Int32un,
    "bits_per_pixel" / Int32un,
    "size" / Int32un,
)

# param_obs_02 has been found to be 90 or 6.
# This first struct handles 90, next handles 6.
# The first "0x1" seems to indicate which type
# of header to expect (0: long, 1: short)
param_obs_02_header = Struct(
    "0x1" / Int16un,
    "0xffff" / Int16un[2],
    "camera_model" / PaddedString(12, "ascii"),
    "image_quality" / PaddedString(24, "ascii"),
    "0x300" / Int16un,
    "0x1" / Int16un,
    "0x0" / Int16un,
    "color_temp" / PaddedString(24, "ascii"),
    "0x2014" / Int16un,
    "zeros" / Int8un[12],
)

param_obs_02_short_header = Struct(
    "0x1" / Int16un,
    "0xffff" / Int16un[2],
)
# phMode = fread(fid, 1, 'uint8');
# phAngle = fread(fid, 1, 'uint8');
# phLightLevel = fread(fid, 1, 'uint16');

img_mot_comp_03_header = Struct(
    "0x0" / Int8un,
    "width" / Int32un,
    "height" / Int32un,
    "bits_per_pixel" / Int32un,
    "num_slices" / Int32un,
)

img_mot_comp_02_header = Struct(
    "temp" / Int8un,
    "motion_width" / Int32un,
    "motion_height" / Int32un,
    "motion_depth" / Int32un,
    "motion_number" / Int32un,
    "motion_format" / Int8un,
    "motion_start_x_pos" / Int32un,
    "motion_start_y_pos" / Int32un,
    "motion_end_x_pos" / Int32un,
    "motion_end_y_pos" / Int32un,
    "size" / Int32un,
)

img_mot_comp_header = Struct(
    "motion_width" / Int32un,
    "motion_height" / Int32un,
    "motion_depth" / Int32un,
    "size" / Int32un,
)

img_projection_header = Struct(
    "width" / Int32un,
    "height" / Int32un,
    "bits_per_pixel" / Int32un,
    "0x1000002" / Int32un,
    "size" / Int32un,
)

effective_scan_range_header = Struct(
    "bounding_box_fundus_pixel" / Int32un[4],
    "bounding_box_trc_pixel" / Int32un[4],
)

regist_info_header = Struct(
    "0x0" / Int8un,
    "unknown" / Int32un[2],
    "bounding_box_in_fundus_pixels" / Int32un[4],
    "version" / PaddedString(32, "ascii"),
    "bounding_box_in_trc_pixels" / Int32un[4],
)

regist_scan_02_header = Struct(
    "clr_scan_region" / Int32un[4],
    "obs_scan_region" / Int32un[4],
)

result_cornea_curve_header = Struct(
    "id" / Int8un[20],
    "width" / Int32un,
    "height" / Int32un,
    "8.0.1.21781" / Int8un[32],
)

result_cornea_thickness_header = Struct(
    "8.0.1.21781" / Int8un[32],
    "id" / Int8un[20],
    "width" / Int32un,
    "height" / Int32un,
)

align_info_header = Struct(
    "unlabeled_1" / Int8un,
    "unlabeled_2" / Int8un,
    "w" / Int32un,
    "n_size" / Int32un,
    "aligndata" / Array(this.w * 2, Int16un),  # if n_size > 0
    # if nblockbytes - (10+n_size) >= 16
    "keyframe_1" / Int32un,
    "keyframe_2" / Int32un,
    "unlabeled_3" / Int32un,
    "unlabeled_4" / Int32un,
)

main_module_info_header = Struct(
    "software_name" / PaddedString(128, "ascii"),
    "file_version_1" / Int16un,
    "file_version_2" / Int16un,
    "file_version_3" / Int16un,
    "file_version_4" / Int16un,
    "string" / PaddedString(128, "ascii"),
)

fast_q2_info_header = Struct("various_quality_statistics" / Float32n[6])

gla_littmann_01_header = Struct("0xffff" / Int32un, "0x1" / Int32un)

thumbnail_header = Struct(
    "size" / Int32un,
    # "img" / Int8un[this.size]
)

patientext_info_header = Struct("unknown" / Int8un[128])



import json

from oct_converter.dicom import create_dicom_from_oct
from oct_converter.readers import FDA

# a sample .fda file can be downloaded from the Biobank resource here:
# https://biobank.ndph.ox.ac.uk/showcase/refer.cgi?id=31
filepath = "/Users/mark/Downloads/eg_oct_fda.fda"
fda = FDA(filepath)

oct_volume = (
    fda.read_oct_volume()
)  # returns an OCT volume with additional metadata if available
oct_volume.peek(show_contours=True)  # plots a montage of the volume
oct_volume.save("fda_testing.avi")  # save volume as a movie
oct_volume.save(
    "fda_testing.png"
)  # save volume as a set of sequential images, fds_testing_[1...N].png

fundus_image = (
    fda.read_fundus_image()
)  # returns a  Fundus image with additional metadata if available
fundus_image.save("fda_testing_fundus.jpg")

fundus_grayscale_image = fda.read_fundus_image_gray_scale()
if fundus_grayscale_image:
    fundus_grayscale_image.save("fda_testing_grayscalefundus.jpg")

# Read segmentation (contours)
segmentation = fda.read_segmentation()

# extract all other metadata
metadata = fda.read_all_metadata()
with open("metadata.json", "w") as outfile:
    outfile.write(json.dumps(metadata, indent=4))

# create a DICOM from FDA
dcm = create_dicom_from_oct(filepath)
# Output dir can be specified, otherwise will
# default to current working directory.



from construct import (
    Array,
    Bytes,
    BytesInteger,
    Computed,
    Float64n,
    Hex,
    Int16un,
    Int32un,
    Lazy,
    PaddedString,
    Seek,
    Struct,
    Tell,
    this,
)

headerField = Struct(
    keylength=Int32un, key=PaddedString(this.keylength, "utf8"), dataLength=Int32un
)

floatField = Struct(
    keylength=Int32un,
    key=PaddedString(this.keylength, "utf8"),
    dataLength=Int32un,
    value=Float64n,
)

intField = Struct(
    keylength=Int32un,
    key=PaddedString(this.keylength, "utf8"),
    dataLength=Int32un,
    value=BytesInteger(this.dataLength, signed=False, swapped=True),
)

lazyIntField = Struct(
    "keylength" / Int32un,
    "key" / PaddedString(this.keylength, "utf8"),
    "dataLength" / Int32un,
    "offset" / Tell,
    "end" / Computed(this.offset + this.dataLength),
    "value" / Lazy(Bytes(this.dataLength)),
    Seek(this.end),
)

date = Struct(
    year=Int16un,
    month=Int16un,
    dow=Int16un,
    day=Int16un,
    hour=Int16un,
    minute=Int16un,
    second=Int16un,
    millisecond=Int16un,
)
dateField = Struct(
    keylength=Int32un,
    key=PaddedString(this.keylength, "utf8"),
    dataLength=Int32un,
    value=date,
)

strField = Struct(
    keylength=Int32un,
    key=PaddedString(this.keylength, "utf8"),
    dataLength=Int32un,
    value=PaddedString(this.dataLength, "utf8"),
)

bioptigen_oct_header_struct = Struct(
    "magicNumber" / Hex(Int32un),
    "version" / Hex(Int16un),
    "frameheader" / headerField,
    "framecount" / intField,
    "linecount" / intField,
    "linelength" / intField,
    "sampleformat" / intField,
    "description" / strField,
    "xmin" / floatField,
    "xmax" / floatField,
    "xcaption" / strField,
    "ymin" / floatField,
    "ymax" / floatField,
    "ycaption" / strField,
    "scantype" / intField,
    "scandepth" / floatField,
    "scanlength" / floatField,
    "azscanlength" / floatField,
    "elscanlength" / floatField,
    "objectdistance" / floatField,
    "scanangle" / floatField,
    "scans" / intField,
    "frames" / intField,
    "dopplerflag" / intField,
    "config" / lazyIntField,
    BytesInteger(4, signed=False, swapped=True),
)

oct_frame_header_struct = Struct(
    "framedata" / headerField,
    "framedatetime" / dateField,
    "frametimestamp" / floatField,
    "framelines" / intField,
    "keylength" / Int32un,
    "key" / PaddedString(this.keylength, "utf8"),
    "dataLength" / Int32un,
)

oct_frame_data_struct = Struct(
    "rows" / Computed(this._._.header.linelength.value),
    "columns" / Computed(this._.header.framelines.value),
    "totalpixels" / Computed(this.rows * this.columns),
    "offset" / Tell,
    "end" / Computed(this.offset + this.totalpixels * 2),
    "pixels" / Lazy(Array(this.totalpixels, Int16un)),
    Seek(this.end),
)
oct_frame_struct = Struct(
    "header" / oct_frame_header_struct,
    "image" / oct_frame_data_struct,
    BytesInteger(4, signed=False, swapped=True),
)
oct_frame_stack_struct = Array(this.header.framecount.value, oct_frame_struct)

bioptigen_file_structure = Struct(
    "header" / bioptigen_oct_header_struct, "data" / oct_frame_stack_struct
)



import json

from oct_converter.dicom import create_dicom_from_oct
from oct_converter.readers import E2E

filepath = "../sample_files/sample.E2E"
file = E2E(filepath)
oct_volumes = (
    file.read_oct_volume()
)  # returns a list of all OCT volumes with additional metadata if available
for volume in oct_volumes:
    volume.peek(show_contours=True)  # plots a montage of the volume
    volume.save("{}_{}.avi".format(volume.volume_id, volume.laterality))

fundus_images = (
    file.read_fundus_image()
)  # returns a list of all fundus images with additional metadata if available
for image in fundus_images:
    image.save("{}+{}.png".format(image.image_id, image.laterality))

# extract all other metadata
metadata = file.read_all_metadata()
with open("metadata.json", "w") as outfile:
    outfile.write(json.dumps(metadata, indent=4))

# create a DICOM from E2E
dcm = create_dicom_from_oct(filepath)
# Output dir can be specified, otherwise will
# default to current working directory.
# If multiple volumes are present in the E2E file,
# multiple files will be created.



from __future__ import annotations

import typing as t
from datetime import datetime
from pathlib import Path

import numpy as np
from construct import ListContainer

from oct_converter.image_types import FundusImageWithMetaData, OCTVolumeWithMetaData
from oct_converter.readers.binary_structs import fds_binary


class FDS(object):
    """Class for extracting data from Topcon's .fds file format.

    Notes:
        Mostly based on description of .fds file format here:
        https://bitbucket.org/uocte/uocte/wiki/Topcon%20File%20Format

    Attributes:
        filepath: path to .img file for reading.
        chunk_dict: names of data chunks present in the file, and their start locations.
    """

    def __init__(self, filepath: str | Path) -> None:
        self.filepath = Path(filepath)
        if not self.filepath.exists():
            raise FileNotFoundError(self.filepath)

        self.chunk_dict, self.header = self.get_list_of_file_chunks()

    def get_list_of_file_chunks(self, printing: bool = False) -> t.Tuple[dict, dict]:
        """Find all data chunks present in the file.

        Returns:
            chunk_dict: dictionary of chunk names, containing their locations
            in the file and size.
            header: dictionary of file header information

        """
        chunk_dict = {}
        with open(self.filepath, "rb") as f:
            raw = f.read(15)
            header = dict(fds_binary.header.parse(raw))

            eof = False
            while not eof:
                chunk_name_size = np.fromstring(f.read(1), dtype=np.uint8)[0]
                if chunk_name_size == 0:
                    eof = True
                else:
                    chunk_name = f.read(chunk_name_size)
                    chunk_size = np.fromstring(f.read(4), dtype=np.uint32)[0]
                    chunk_location = f.tell()
                    f.seek(chunk_size, 1)
                    chunk_dict[chunk_name] = [chunk_location, chunk_size]
        if printing:
            print("File {} contains the following chunks:".format(self.filepath))
            for key in chunk_dict.keys():
                print(key)
            print("")
        return chunk_dict, header

    def read_oct_volume(self) -> OCTVolumeWithMetaData:
        """Reads OCT data.

        Returns:
            OCTVolumeWithMetaData
        """
        # TODO: could support the other IMG_SCAN variants as well.
        if b"@IMG_SCAN_03" not in self.chunk_dict:
            raise ValueError("Could not find OCT header @IMG_SCAN_03 in chunk list")
        with open(self.filepath, "rb") as f:
            chunk_location, chunk_size = self.chunk_dict[b"@IMG_SCAN_03"]
            f.seek(chunk_location)
            raw = f.read(22)
            oct_header = fds_binary.oct_header.parse(raw)
            number_pixels = (
                oct_header.width * oct_header.height * oct_header.number_slices
            )
            raw_volume = np.fromstring(f.read(number_pixels * 2), dtype=np.uint16)
            volume = np.array(raw_volume)
            volume = volume.reshape(
                oct_header.width, oct_header.height, oct_header.number_slices, order="F"
            )
            volume = np.transpose(volume, [1, 0, 2])

        # calculate pixel spacing
        pixel_spacing = self.read_scan_params(oct_header)

        # read all other metadata
        metadata = self.read_all_metadata()
        patient_info = metadata.get("patient_info_02") or metadata.get(
            "patient_info", {}
        )
        capture_info = metadata.get("capture_info_02") or metadata.get(
            "capture_info", {}
        )
        sex_map = {1: "M", 2: "F", 3: "O", None: ""}
        lat_map = {0: "R", 1: "L", None: ""}

        try:
            patient_dob = datetime(*patient_info.get("birth_date"))
        except (TypeError, ValueError):
            patient_dob = None

        oct_volume = OCTVolumeWithMetaData(
            [volume[:, :, i] for i in range(volume.shape[2])],
            patient_id=patient_info.get("patient_id"),
            first_name=patient_info.get("first_name"),
            surname=patient_info.get("last_name"),
            sex=sex_map[patient_info.get("sex", None)],
            patient_dob=patient_dob,
            acquisition_date=datetime(*capture_info.get("cap_date")),
            laterality=lat_map[capture_info.get("eye", None)],
            pixel_spacing=pixel_spacing,
            metadata=metadata,
            header=self.header,
            oct_header=dict(oct_header),
        )
        return oct_volume

    def read_fundus_image(self) -> FundusImageWithMetaData:
        """Reads fundus image.

        Returns:
            FundusImageWithMetaData
        """
        if b"@IMG_OBS" not in self.chunk_dict:
            raise ValueError("Could not find fundus header @IMG_OBS in chunk list")
        with open(self.filepath, "rb") as f:
            chunk_location, chunk_size = self.chunk_dict[b"@IMG_OBS"]
            f.seek(chunk_location)
            raw = f.read(21)
            fundus_header = fds_binary.fundus_header.parse(raw)
            # number_pixels = fundus_header.width * fundus_header.height * fundus_header.number_slices
            raw_image = np.fromstring(f.read(fundus_header.size), dtype=np.uint8)
            # raw_image = [struct.unpack('B', f.read(1)) for pixel in range(fundus_header.size)]
            image = np.array(raw_image)
            image = image.reshape(
                3, fundus_header.width, fundus_header.height, order="F"
            )
            image = np.transpose(image, [2, 1, 0])
            image = image.astype(np.float32)
            # store with RGB channel order
            image = np.flip(image, 2)
        fundus_image = FundusImageWithMetaData(image)
        return fundus_image

    def read_scan_params(self, oct_header: dict) -> list:
        """Given available chunks, identifies available PARAM_SCAN chunk
        and calculates pixel spacing.

        Args:
            oct_header: OCT header information as dict

        Returns:
            pixel_spacing: list of pixel spacing, ordered by width, slice thickness, height.
        """
        if (
            b"@PARAM_SCAN_04" not in self.chunk_dict
            and b"@PARAM_SCAN_02" not in self.chunk_dict
        ):
            print(
                "Neither @PARAM_SCAN_04 nor @PARAM_SCAN_02 found. Pixel spacing not calculated."
            )
            return None
        with open(self.filepath, "rb") as f:
            if b"@PARAM_SCAN_04" in self.chunk_dict:
                chunk_loc, chunk_size = self.chunk_dict.get(
                    b"@PARAM_SCAN_04", (None, None)
                )
                f.seek(chunk_loc)
                scan_params = fds_binary.param_scan_04_header.parse(f.read(chunk_size))
            elif b"@PARAM_SCAN_02" in self.chunk_dict:
                chunk_loc, chunk_size = self.chunk_dict.get(
                    b"@PARAM_SCAN_02", (None, None)
                )
                f.seek(chunk_loc)
                scan_params = fds_binary.param_scan_02_header.parse(f.read(chunk_size))
            pixel_spacing = [
                scan_params.get("x_dimension_mm")
                / oct_header.get("width"),  # WidthPixelS, PixelSpacing[1]
                scan_params.get("y_dimension_mm")
                / oct_header.get("number_slices"),  # FramePixelS / SliceThickness
                scan_params.get("z_resolution_um")
                / 1000,  # zHeightPixelS, PixelSpacing[0]
            ]
        return pixel_spacing

    def read_all_metadata(self, verbose: bool = False):
        """
        Reads all available metadata and returns a dictionary.

        Args:
            verbose: if True, prints the chunks that are not supported.

        Returns:
            dictionary with all metadata.
        """
        metadata = dict()
        for key in self.chunk_dict.keys():
            if key in [b"IMG_SCAN_03", b"@IMG_OBS"]:
                # these chunks have their own dedicated methods for extraction
                continue
            elif key == b"@PARAM_OBS_02":
                metadata["param_obs_02"] = self.read_param_obs()
                continue
            json_key = key.decode().split("@")[-1].lower()
            try:
                metadata[json_key] = self.read_any_info_and_make_dict(key)
            except KeyError:
                if verbose:
                    print(f"{key} there is no method for getting info from this chunk.")
        return metadata

    def read_any_info_and_make_dict(self, chunk_name: str) -> dict:
        """Reads any chunk and constructs a dictionary.

        Args:
            chunk_name: name of the chunk which data will be taken.

        Returns:
            Chunk info data
        """
        if chunk_name not in self.chunk_dict:
            print(f"{chunk_name} is not in chunk list, skipping.")
            return None
        with open(self.filepath, "rb") as f:
            chunk_location, chunk_size = self.chunk_dict[chunk_name]
            f.seek(chunk_location)  # Set the chunk’s current position.
            raw = f.read()
            header_name = f"{chunk_name.decode().split('@')[-1].lower()}_header"
            chunk_info_header = dict(fds_binary.__dict__[header_name].parse(raw))
            chunks_info = dict()
            for idx, key in enumerate(chunk_info_header.keys()):
                if idx == 0:
                    continue
                if type(chunk_info_header[key]) is ListContainer:
                    chunks_info[key] = list(chunk_info_header[key])
                else:
                    chunks_info[key] = chunk_info_header[key]
        return chunks_info

    def read_param_obs(self) -> dict:
        """Reads PARAM_OBS_02 while accounting for varied chunk sizes.

        Returns:
            Chunk info data for PARAM_OBS_02
        """
        with open(self.filepath, "rb") as f:
            chunk_location, chunk_size = self.chunk_dict[b"@PARAM_OBS_02"]
            f.seek(chunk_location)  # Set the chunk’s current position.
            raw = f.read()
            # PARAM_OBS_02 is either of size 90 or size 6.
            if chunk_size == 90:
                chunk_info_header = dict(fds_binary.param_obs_02_header.parse(raw))
            else:  # chunk_size == 6
                chunk_info_header = dict(
                    fds_binary.param_obs_02_short_header.parse(raw)
                )

            chunks_info = dict()
            for idx, key in enumerate(chunk_info_header.keys()):
                if idx == 0:
                    continue
                if type(chunk_info_header[key]) is ListContainer:
                    chunks_info[key] = list(chunk_info_header[key])
                else:
                    chunks_info[key] = chunk_info_header[key]
        return chunks_info



from __future__ import annotations

from datetime import datetime
from pathlib import Path

import cv2
import imageio
import matplotlib.pyplot as plt
import numpy as np

VIDEO_TYPES = [
    ".avi",
    ".mp4",
]
IMAGE_TYPES = [".png", ".bmp", ".tiff", ".jpg", ".jpeg"]


class OCTVolumeWithMetaData(object):
    """Class to hold an OCT volume and any related metadata.

    Also provides methods for viewing and saving.

    Attributes:
        volume: all the volume's b-scans.

        patient_id: patient ID.
        first_name: patient first name.
        surname: patient second name.
        sex: patient sex.
        DOB: patient date of birth.

        volume_id: volume ID.
        acquisition_date: date image acquired.
        num_slices: number of b-scans present in volume.
        laterality: left or right eye.

        contours: contours data.
        pixel_spacing: (x, y, z) pixel spacing in mm.
        metadata: all metadata available in the OCT scan.
    """

    def __init__(
        self,
        volume: list[np.array],
        patient_id: str | None = None,
        first_name: str | None = None,
        surname: str | None = None,
        sex: str | None = None,
        patient_dob: str | None = None,
        volume_id: str | None = None,
        acquisition_date: datetime | None = None,
        laterality: str | None = None,
        contours: dict | None = None,
        pixel_spacing: list[float] | None = None,
        metadata: dict | None = None,
        header: dict | None = None,
        oct_header: dict | None = None,
    ) -> None:
        # image
        self.volume = volume

        # patient data
        self.patient_id = patient_id
        self.first_name = first_name
        self.surname = surname
        self.sex = sex
        self.DOB = patient_dob

        # volume data
        self.volume_id = volume_id
        self.acquisition_date = acquisition_date
        self.laterality = laterality
        self.num_slices = len(self.volume)
        self.contours = contours

        # geom data
        self.pixel_spacing = pixel_spacing

        # metadata
        self.metadata = metadata
        self.header = header
        self.oct_header = oct_header

    def peek(
        self,
        rows: int = 5,
        cols: int = 5,
        filepath: str | Path | None = None,
        show_contours: bool | None = False,
    ) -> None:
        """Plots a montage of the OCT volume. Optionally saves the plot if a filepath is provided.

        Args:
            rows: number of rows in the plot.
            cols: number of columns in the plot.
            filepath: location to save montage to.
            show_contours: if set to ``True``, will plot contours on the OCT volume.
        """
        images = rows * cols
        x_size = rows * self.volume[0].shape[0]
        y_size = cols * self.volume[0].shape[1]
        ratio = y_size / x_size
        slices_indices = np.linspace(0, self.num_slices - 1, images).astype(np.int16)
        plt.figure(figsize=(12 * ratio, 12))
        for i in range(images):
            slice_id = slices_indices[i]
            plt.subplot(rows, cols, i + 1)
            plt.imshow(self.volume[slice_id], cmap="gray")
            if show_contours and self.contours is not None:
                for v in self.contours.values():
                    if (
                        slice_id < len(v)
                        and v[slice_id] is not None
                        and not np.isnan(v[slice_id]).all()
                    ):
                        plt.plot(v[slice_id], color="r")
            plt.axis("off")
            plt.title("{}".format(slice_id))
        plt.suptitle("OCT volume with {} slices.".format(self.num_slices))

        if filepath is not None:
            plt.savefig(filepath)
        else:
            plt.show()

    def save(self, filepath: str | Path) -> None:
        """Saves OCT volume as a video or stack of slices.

        Args:
            filepath: location to save volume to. Extension must be in VIDEO_TYPES or IMAGE_TYPES.
        """
        extension = Path(filepath).suffix
        if extension.lower() in VIDEO_TYPES:
            video_writer = imageio.get_writer(filepath, macro_block_size=None)
            for slice in self.volume:
                slice = slice.astype("uint8")
                video_writer.append_data(slice)
            video_writer.close()
        elif extension.lower() in IMAGE_TYPES:
            base = Path(filepath).stem
            print(
                "Saving OCT as sequential slices {}_[1..{}]{}".format(
                    base, len(self.volume), extension
                )
            )
            full_base = Path(filepath).with_suffix("")
            self.volume = np.array(self.volume).astype("float64")
            self.volume *= 255.0 / self.volume.max()
            for index, slice in enumerate(self.volume):
                filename = "{}_{}{}".format(full_base, index, extension)
                cv2.imwrite(filename, slice)
        elif extension.lower() == ".npy":
            np.save(filepath, self.volume)
        else:
            raise NotImplementedError(
                "Saving with file extension {} not supported".format(extension)
            )

    def get_projection(self) -> np.array:
        """Produces a 2D projection image from the volume."""
        projection = np.mean(self.volume, axis=1)
        return projection

    def save_projection(self, filepath: str | Path) -> None:
        """Save a 2D projection image from the volume.

        Args:
            filepath: location to save volume to. Extension must be in IMAGE_TYPES.
        """
        extension = Path(filepath).suffix
        if extension.lower() in IMAGE_TYPES:
            projection = self.get_projection()
            projection = 255 * projection / projection.max()
            cv2.imwrite(filepath, projection.astype(int))
        else:
            raise NotImplementedError(
                "Saving with file extension {} not supported".format(extension)
            )



from __future__ import annotations

from pathlib import Path

import cv2
import numpy as np

VIDEO_TYPES = [
    ".avi",
    ".mp4",
]
IMAGE_TYPES = [".png", ".bmp", ".tiff", ".jpg", ".jpeg"]


class FundusImageWithMetaData(object):
    """Class to hold a fundus image and any related metadata.

    Also provides methods for viewing and saving.

    Attributes:
        image: fundus image.
        laterality: left or right eye.
        patient_id: patient ID.
        image_id: image ID.
        DOB: patient date of birth.
    """

    def __init__(
        self,
        image: np.array,
        laterality: str | None = None,
        patient_id: str | None = None,
        image_id: str | None = None,
        patient_dob: str | None = None,
        metadata: dict | None = None,
    ) -> None:
        self.image = image
        self.laterality = laterality
        self.patient_id = patient_id
        self.image_id = image_id
        self.DOB = patient_dob
        self.metadata = metadata

    def save(self, filepath: str | Path) -> None:
        """Saves fundus image.

        Args:
            filepath: location to save volume to. Extension must be in IMAGE_TYPES.
        """
        extension = Path(filepath).suffix
        if extension.lower() in IMAGE_TYPES:
            # change channel order from RGB to BGR and save with cv2
            image = cv2.cvtColor(self.image, cv2.COLOR_RGB2BGR)
            cv2.imwrite(filepath, image)
        elif extension.lower() == ".npy":
            np.save(filepath, self.image)
        else:
            raise NotImplementedError(
                "Saving with file extension {} not supported".format(extension)
            )



from __future__ import annotations

import re
import tempfile
from datetime import datetime
from pathlib import Path
from typing import BinaryIO

import h5py
import numpy as np
from construct import Struct
from numpy.typing import NDArray

from oct_converter.exceptions import InvalidOCTReaderError
from oct_converter.image_types import OCTVolumeWithMetaData
from oct_converter.readers.binary_structs import boct_binary


class BOCT(object):
    """Class for extracting data from Bioptigen's .OCT file format.

    .OCT stores 4D volumes (time series of 3D volumes with the same shape)

    Attributes:
        filepath: path to .img file for reading.
        file_structure: defines structure of volume's header.
    """

    bioptigen_scan_type_map = {0: "linear", 1: "rect", 3: "rad"}
    file_structure = boct_binary.bioptigen_file_structure
    header_structure = boct_binary.bioptigen_oct_header_struct

    def __init__(self, filepath: Path | str) -> None:
        self.filepath = Path(filepath)
        if not self.filepath.exists():
            raise FileNotFoundError(self.filepath)
        self._validate(self.filepath)

    def _validate(self, path: Path) -> bool:
        try:
            self.header_structure.parse_file(path)
        except UnicodeDecodeError:
            raise InvalidOCTReaderError(
                "OCT header does not match Bioptigen .OCT format. Did you mean to use Optovue .oct (POCT)?"
            )
        return True

    def read_oct_volume(
        self, diskbuffered: bool = False
    ) -> list[OCTVolumeWithMetaData]:
        """Reads OCT data.

        Args:
            diskbuffered: if True, reduces memory usage by storing volume on disk using HDF5.

        Returns:
            OCT volumes with metadata.
        """
        # Laterality/patient_id data not contained in .OCT file (often in filename)
        self.laterality = None
        self.patient_id = self.filepath.stem

        # Lazily parse the file without loading frame pixels
        oct = self.file_structure.parse_file(self.filepath)
        header = oct.header
        self.frames = FrameGenerator(oct.data)
        scantype = self.bioptigen_scan_type_map[header.scantype.value]
        framecount = header.frames.value
        scancount = header.scans.value
        if scantype == "linear":
            # linear bscans can contain multiple scans at one position
            # reorder into (framecount*scancount,1,y,x)
            framecount *= scancount
            scancount = 1

        self.volume_shape = (
            framecount,
            scancount,
            self.frames.Ascans,
            self.frames.depth,
        )
        bscan_shape = (self.volume_shape[2], self.volume_shape[3])
        self.vol_frames_shape = (self.volume_shape[0], self.volume_shape[1])
        if diskbuffered:
            self.vol = self._create_disk_buffer(buffer_shape=bscan_shape)
        else:
            self.vol = np.empty(self.volume_shape, dtype=np.uint16)

        # Grab the acquisition datetime,
        dt = oct.data[0].header.framedatetime.value
        self.acquisition_datetime = datetime(
            year=dt.year,
            month=dt.month,
            day=dt.day,
            hour=dt.hour,
            minute=dt.minute,
            second=dt.second,
        )

        # Attempt to parse laterality from filename
        self.get_laterality_from_filename()

        return self.load_oct_volume()

    def _create_disk_buffer(
        self, buffer_shape: tuple[int, int], name: str = "vol"
    ) -> h5py.Dataset:
        x, y = buffer_shape
        chunksize = (1, 1, x, y)
        tf = h5py.File(tempfile.TemporaryFile(), "w")
        return tf.create_dataset(
            name, shape=self.volume_shape, dtype=np.uint16, chunks=chunksize
        )

    def load_oct_volume(self) -> list[OCTVolumeWithMetaData]:
        volFrames = np.reshape(self.frames.data, self.vol_frames_shape)
        try:
            with open(self.filepath, "rb") as f:
                for t, v in enumerate(volFrames):
                    for z, frame in enumerate(v):
                        self.vol[t, z, :, :] = frame.load(
                            f, self.frames.Ascans, self.frames.depth
                        )
        except Exception as e:
            print(e)
            print("Stopping load")
        return [
            OCTVolumeWithMetaData(
                self.vol[t, :, :, :],
                acquisition_date=self.acquisition_datetime,
                laterality=self.laterality,
            )
            for t in range(self.vol.shape[0])
        ]

    def read_fundus_image(self) -> None:
        return

    def get_laterality_from_filename(self) -> None:
        """Attempts to find laterality within the filename,
        if found, sets self.laterality

        Returns:
            None
        """
        filename = Path(self.filepath).name

        lat_from_filename = (
            re.search(r"O[D|S]", filename).group(0)
            if re.search(r"O[D|S]", filename)
            else None
        )

        if lat_from_filename == "OD":
            self.laterality = "R"
        elif lat_from_filename == "OS":
            self.laterality = "L"
        else:
            self.laterality = ""

        return


class OCTFrame(object):
    def __init__(self, frame: Struct) -> None:
        self.count = frame.image.totalpixels
        self.abs_pos = frame.image.offset

    def from_bytes(self, f: BinaryIO) -> NDArray[np.uint16]:
        f.seek(self.abs_pos, 0)
        im = np.fromfile(f, dtype=np.uint16, count=self.count)
        return im

    def load(self, f: BinaryIO, Ascans: int, depth: int) -> NDArray[np.uint16]:
        return np.resize(self.from_bytes(f), (Ascans, depth))


class FrameGenerator(object):
    def __init__(self, oct_data: Struct) -> None:
        self.Ascans = oct_data[0].image.columns
        self.depth = oct_data[0].image.rows
        self.data = np.asarray([OCTFrame(frame) for frame in oct_data])
        self.count = len(self.data)

    def reorder(self, indexArr: NDArray[np.int_]) -> FrameGenerator:
        try:
            self.data = self.data[indexArr]
        except Exception as e:
            print(e)
        finally:
            return self



from __future__ import annotations

import re
from datetime import datetime
from pathlib import Path

import numpy as np

from oct_converter.image_types import OCTVolumeWithMetaData


class IMG(object):
    """Class for extracting data from Zeiss's .img file format.

    Attributes:
        filepath: path to .img file for reading.
    """

    def __init__(self, filepath: str | Path) -> None:
        self.filepath = Path(filepath)
        if not self.filepath.exists():
            raise FileNotFoundError(self.filepath)

    def read_oct_volume(
        self, rows: int = 1024, cols: int = 512, interlaced: bool = False
    ):
        """Reads OCT data.

        Args:
            rows: can be used to specify a custom row dimension of the image slice. Defaults to 1024 pixels.
            cols: dan be used to specify a custom column dimension of the image slice. Defaults to 512 pixels.
            interlaced: determines whether data needs to be de-interlaced.

        Returns:
            OCTVolumeWithMetaData
        """
        with open(self.filepath, "rb") as f:
            volume = np.frombuffer(
                f.read(), dtype=np.uint8
            )  # np.fromstring() gives numpy depreciation warning
            num_slices = len(volume) // (rows * cols)
            volume = volume.reshape((rows, cols, num_slices), order="F")
            if interlaced:
                shape = volume.shape
                interlaced = np.zeros((int(shape[0] / 2), shape[1], shape[2] * 2))
                interlaced[..., 0::2] = volume[:cols, ...]
                interlaced[..., 1::2] = volume[cols:, ...]
                interlaced = np.rot90(interlaced, axes=(0, 1))
                volume = interlaced

        meta = self.get_metadata_from_filename()
        lat_map = {"OD": "R", "OS": "L", None: ""}

        oct_volume = OCTVolumeWithMetaData(
            [volume[:, :, i] for i in range(volume.shape[2])],
            patient_id=meta.get("patient_id"),
            acquisition_date=meta.get("acquisition_date"),
            laterality=lat_map[meta.get("laterality", None)],
            metadata=meta,
        )
        return oct_volume

    def get_metadata_from_filename(self) -> dict:
        """Attempts to find metadata within the filename

        Returns:
            meta: dict of information extracted from filename
        """
        filename = Path(self.filepath).name
        meta = {}
        meta["patient_id"] = (
            re.search(r"^P\d+", filename).group(0)
            if re.search(r"^P\d+", filename)
            else None
        )
        acq = (
            list(
                re.search(
                    r"(?P<m>\d{1,2})-(?P<d>\d{1,2})-(?P<y>\d{4})_(?P<h>\d{1,2})-(?P<M>\d{1,2})-(?P<s>\d{1,2})",
                    filename,
                ).groups()
            )
            if re.search(
                r"(?P<m>\d{1,2})-(?P<d>\d{1,2})-(?P<y>\d{4})_(?P<h>\d{1,2})-(?P<M>\d{1,2})-(?P<s>\d{1,2})",
                filename,
            )
            else None
        )
        if acq:
            meta["acquisition_date"] = datetime(
                year=int(acq[2]),
                month=int(acq[0]),
                day=int(acq[1]),
                hour=int(acq[3]),
                minute=int(acq[4]),
                second=int(acq[5]),
            )
        meta["laterality"] = (
            re.search(r"O[D|S]", filename).group(0)
            if re.search(r"O[D|S]", filename)
            else None
        )
        meta["sn"] = (
            re.search(r"sn\d+", filename).group(0)
            if re.search(r"sn\d*", filename)
            else None
        )

        return meta



from construct import (
    Array,
    Float32l,
    Int8un,
    Int16un,
    Int32sn,
    Int32un,
    Int64un,
    PaddedString,
    Struct,
)

# Mostly based on description of .e2e file format here:
#         https://bitbucket.org/uocte/uocte/wiki/Heidelberg%20File%20Format.

header_structure = Struct(
    "magic1" / PaddedString(12, "ascii"),
    "version" / Int32un,
    "unknown" / Array(10, Int16un),
)
main_directory_structure = Struct(
    "magic2" / PaddedString(12, "ascii"),
    "version" / Int32un,
    "unknown" / Array(10, Int16un),
    "num_entries" / Int32un,
    "current" / Int32un,
    "prev" / Int32un,
    "unknown3" / Int32un,
)
sub_directory_structure = Struct(
    "pos" / Int32un,
    "start" / Int32un,
    "size" / Int32un,
    "unknown" / Int32un,
    # Patient DB ID is set by the software
    # and is not necessarily equal to patient_id_structure's patient_id
    "patient_db_id" / Int32un,
    "study_id" / Int32un,
    "series_id" / Int32un,
    "slice_id" / Int32sn,
    "unknown2" / Int16un,
    "unknown3" / Int16un,
    "type" / Int32un,
    "unknown4" / Int32un,
)
chunk_structure = Struct(
    "magic3" / PaddedString(12, "ascii"),
    "unknown" / Int32un,
    "unknown2" / Int32un,
    "pos" / Int32un,
    "size" / Int32un,
    "unknown3" / Int32un,
    # Patient DB ID is set by the software
    # and is not necessarily equal to patient_id_structure's patient_id
    "patient_db_id" / Int32un,
    "study_id" / Int32un,
    "series_id" / Int32un,
    "slice_id" / Int32sn,
    "ind" / Int16un,
    "unknown4" / Int16un,
    "type" / Int32un,
    "unknown5" / Int32un,
)
image_structure = Struct(
    "size" / Int32un,
    "type" / Int32un,
    "unknown" / Int32un,
    "height" / Int32un,
    "width" / Int32un,
)
patient_id_structure = Struct(
    "first_name" / PaddedString(31, "ascii"),
    "surname" / PaddedString(51, "ascii"),
    "title" / PaddedString(15, "ascii"),
    "birthdate" / Int32un,
    "sex" / PaddedString(1, "ascii"),
    "patient_id" / PaddedString(25, "ascii"),
)
lat_structure = Struct(
    "unknown" / Array(14, Int8un), "laterality" / Int8un, "unknown2" / Int8un
)
contour_structure = Struct(
    "unknown0" / Int32un,
    "id" / Int32un,
    "unknown1" / Int32un,
    "width" / Int32un,
)

# following the spec from
# https://github.com/neurodial/LibE2E/blob/d26d2d9db64c5f765c0241ecc22177bb0c440c87/E2E/dataelements/bscanmetadataelement.cpp#L75
bscan_metadata = Struct(
    "unknown1" / Int32un,
    "imgSizeY" / Int32un,
    "imgSizeX" / Int32un,
    "posX1" / Float32l,
    "posY1" / Float32l,
    "posX2" / Float32l,
    "posY2" / Float32l,
    "zero1" / Int32un,
    "unknown2" / Float32l,
    "scaley" / Float32l,
    "unknown3" / Float32l,
    "zero2" / Int32un,
    "unknown4" / Array(2, Float32l),
    "zero3" / Int32un,
    "imgSizeWidth" / Int32un,
    "numImages" / Int32un,
    "aktImage" / Int32un,
    "scanType" / Int32un,
    "centrePosX" / Float32l,
    "centrePosY" / Float32l,
    "unknown5" / Int32un,
    "acquisitionTime" / Int64un,
    "numAve" / Int32un,
    "imgQuality" / Float32l,
)



import dataclasses
import datetime
import enum
import typing as t
from dataclasses import field


class OPTAcquisitionDevice(enum.Enum):
    """OPT Acquisition Device enumeration.

    Contains code designator, code value and code meaning for each entry.
    """

    OCTScanner = ("SRT", "A-00FBE", "Optical Coherence Tomography Scanner")
    RetinalThicknessAnalyzer = ("SRT", "R-FAB5A", "Retinal Thickness Analyzer")
    ConfocalScanningLaserOphthalmoscope = (
        "SRT",
        "A-00E8B",
        "Confocal Scanning Laser Ophthalmoscope",
    )
    ScheimpflugCamera = ("DCM", "111626", "Scheimpflug Camera")
    ScanningLaserPolarimeter = ("SRT", "A-00E8C", "Scanning Laser Polarimeter")
    ElevationBasedCornealTomographer = (
        "DCM",
        "111945",
        "Elevation-based corneal tomographer",
    )
    ReflectionBasedCornealTopographer = (
        "DCM",
        "111946",
        "Reflection-based corneal topographer",
    )
    InterferometryBasedCornealTomographer = (
        "DCM",
        "111947",
        "Interferometry-based corneal tomographer",
    )
    Unspecified = ("OCT-converter", "D-0001", "Unspecified scanner")


class OPTAnatomyStructure(enum.Enum):
    """OPT Anatomy enumeration.

    Contains code designator, code value and code meaning for each entry.
    """

    AnteriorChamberOfEye = ("SRT", "T-AA050", "Anterior chamber of eye")
    BothEyes = ("SRT", "T-AA180", "Both eyes")
    ChoroidOfEye = ("SRT", "T-AA310", "Choroid of eye")
    CiliaryBody = ("SRT", "T-AA400", "Ciliary body")
    Conjunctiva = ("SRT", "T-AA860", "Conjunctiva")
    Cornea = ("SRT", "T-AA200", "Cornea")
    Eye = ("SRT", "T-AA000", "Eye")
    Eyelid = ("SRT", "T-AA810", "Eyelid")
    FoveaCentralis = ("SRT", "T-AA621", "Fovea centralis")
    Iris = ("SRT", "T-AA500", "Iris")
    LacrimalCaruncle = ("SRT", "T-AA862", "Lacrimal caruncle")
    LacrimalGland = ("SRT", "T-AA910", "Lacrimal gland")
    LacrimalSac = ("SRT", "T-AA940", "Lacrimal sac")
    Lens = ("SRT", "T-AA700", "Lens")
    LowerEyeLid = ("SRT", "T-AA830", "Lower Eyelid")
    OphthalmicArtery = ("SRT", "T-45400", "Ophthalmic artery")
    OpticNerveHead = ("SRT", "T-AA630", "Optic nerve head")
    Retina = ("SRT", "T-AA610", "Retina")
    Sclera = ("SRT", "T-AA110", "Sclera")
    UpperEyeLid = ("SRT", "T-AA820", "Upper Eyelid")
    Unspecified = ("OCT-converter", "A-0001", "Unspecified anatomy")


class OCTDetectorType(enum.Enum):
    CCD = "CCD"
    CMOS = "CMOS"
    PHOTO = "PHOTO"
    INT = "INT"
    Unknown = "UNKNOWN"


@dataclasses.dataclass
class PatientMeta:
    # Patient Info
    first_name: str = ""
    last_name: str = ""
    patient_id: str = ""
    patient_sex: str = ""
    patient_dob: t.Optional[datetime.datetime] = None


@dataclasses.dataclass
class SeriesMeta:
    # Study and Series
    study_id: str = ""
    series_id: str = ""
    laterality: str = ""
    acquisition_date: t.Optional[datetime.datetime] = None
    # Anatomy
    opt_anatomy: OPTAnatomyStructure = OPTAnatomyStructure.Unspecified


@dataclasses.dataclass
class ManufacturerMeta:
    # Manufacturer info
    manufacturer: str = ""
    manufacturer_model: str = "unknown"
    device_serial: str = "unknown"
    software_version: str = "unknown"


@dataclasses.dataclass
class ImageGeometry:
    # Image geometry info
    pixel_spacing: list[float] = field(default_factory=list)
    slice_thickness: float = 1.0
    image_orientation: list[float] = field(default_factory=list)


@dataclasses.dataclass
class OCTImageParams:
    # PS3.3 C.8.17.9
    opt_acquisition_device: OPTAcquisitionDevice = OPTAcquisitionDevice.Unspecified
    DetectorType: OCTDetectorType = OCTDetectorType.Unknown
    IlluminationWaveLength: t.Optional[float] = None
    IlluminationPower: t.Optional[float] = None
    IlluminationBandwidth: t.Optional[float] = None
    DepthSpatialResolution: t.Optional[float] = None
    MaximumDepthDistortion: t.Optional[float] = None
    AlongscanSpatialResolution: t.Optional[float] = None
    MaximumAlongscanDistortion: t.Optional[float] = None
    AcrossscanSpatialResolution: t.Optional[float] = None
    MaximumAcrossscanDistortion: t.Optional[float] = None

    # NOTE: Could eventually include C.8.17.8 Acquisition Params.


@dataclasses.dataclass
class DicomMetadata:
    patient_info: PatientMeta
    series_info: SeriesMeta
    manufacturer_info: ManufacturerMeta

    image_geometry: ImageGeometry
    oct_image_params: OCTImageParams



from __future__ import annotations

from oct_converter.dicom.metadata import (
    DicomMetadata,
    ImageGeometry,
    ManufacturerMeta,
    OCTDetectorType,
    OCTImageParams,
    OPTAcquisitionDevice,
    OPTAnatomyStructure,
    PatientMeta,
    SeriesMeta,
)
from oct_converter.image_types import FundusImageWithMetaData, OCTVolumeWithMetaData


def e2e_patient_meta(meta: dict) -> PatientMeta:
    """Creates PatientMeta from e2e info stored in raw metadata

    Args:
        meta: Nested dictionary of metadata accumulated by the E2E reader
    Returns:
        PatientMeta: Patient metadata populated by oct
    """
    patient = PatientMeta()

    patient_data = meta.get("patient_data", [{}])

    patient.first_name = patient_data[0].get("first_name")
    patient.last_name = patient_data[0].get("surname")
    patient.patient_id = patient_data[0].get("patient_id")
    patient.patient_sex = patient_data[0].get("sex")
    # TODO patient.patient_dob
    # Currently, E2E's patient_dob is incorrect, see
    # the E2E reader for more context.

    return patient


def e2e_series_meta(id, laterality, acquisition_date) -> SeriesMeta:
    """Creates SeriesMeta from info parsed by the E2E reader

    Args:
        id: Equivalent to oct.volume_id or fundus.image_id
        laterality: R or L, from image.laterality
        acquisition_date: Scan date for OCT, or None for fundus
    Returns:
        SeriesMeta: Series metadata populated by oct
    """
    patient_db_id, study_id, series_id = id.split("_")
    series = SeriesMeta()

    series.study_id = study_id
    series.series_id = series_id
    series.laterality = laterality
    series.acquisition_date = acquisition_date
    series.opt_anatomy = OPTAnatomyStructure.Retina

    return series


def e2e_manu_meta() -> ManufacturerMeta:
    """Creates ManufacturerMeta with Heidelberg defaults.

    Args:
        None
    Returns:
        ManufacturerMeta: Manufacture metadata module
    """
    manufacture = ManufacturerMeta()

    manufacture.manufacturer = "Heidelberg Engineering"
    manufacture.manufacturer_model = "Spectralis"
    manufacture.device_serial = ""
    manufacture.software_version = ""

    return manufacture


def e2e_image_geom(pixel_spacing: list) -> ImageGeometry:
    """Creates ImageGeometry from E2E metadata

    Args:
        pixel_spacing: Pixel spacing identified by E2E reader
    Returns:
        ImageGeometry: Geometry data populated by pixel_spacing
    """
    image_geom = ImageGeometry()
    image_geom.pixel_spacing = [pixel_spacing[1], pixel_spacing[0]]
    image_geom.slice_thickness = pixel_spacing[2]
    image_geom.image_orientation = [1, 0, 0, 0, 1, 0]

    return image_geom


def e2e_image_params() -> OCTImageParams:
    """Creates OCTImageParams specific to E2E

    Args:
        None
    Returns:
        OCTImageParams: Image params populated with E2E defaults
    """
    image_params = OCTImageParams()
    image_params.opt_acquisition_device = OPTAcquisitionDevice.OCTScanner
    image_params.DetectorType = OCTDetectorType.CCD
    image_params.IlluminationWaveLength = 880
    image_params.IlluminationPower = 1200
    image_params.IlluminationBandwidth = 50
    image_params.DepthSpatialResolution = 7
    image_params.MaximumDepthDistortion = 0.5
    image_params.AlongscanSpatialResolution = 13
    image_params.MaximumAlongscanDistortion = 0.5
    image_params.AcrossscanSpatialResolution = 13
    image_params.MaximumAcrossscanDistortion = 0.5

    return image_params


def e2e_dicom_metadata(
    image: FundusImageWithMetaData | OCTVolumeWithMetaData,
) -> DicomMetadata:
    """Creates DicomMetadata for oct or fundus image and populates each module

    Args:
        image: Oct or Fundus image type created by the E2E reader
    Returns:
        DicomMetadata: Populated DicomMetadata created with fundus or oct metadata
    """

    meta = DicomMetadata
    meta.patient_info = e2e_patient_meta(image.metadata)
    meta.manufacturer_info = e2e_manu_meta()
    meta.oct_image_params = e2e_image_params()
    if type(image) == OCTVolumeWithMetaData:
        meta.series_info = e2e_series_meta(
            image.volume_id, image.laterality, image.acquisition_date
        )
        meta.image_geometry = e2e_image_geom(image.pixel_spacing)
    else:  # type(image) == FundusImageWithMetaData
        meta.series_info = e2e_series_meta(image.image_id, image.laterality, None)
        meta.image_geometry = e2e_image_geom([1, 1, 1])

    return meta



import typing as t
from datetime import datetime
from importlib import metadata
from pathlib import Path

import numpy as np
from construct import StreamError
from pydicom.dataset import Dataset, FileDataset, FileMetaDataset
from pydicom.uid import (
    ExplicitVRLittleEndian,
    OphthalmicTomographyImageStorage,
    generate_uid,
)

from oct_converter.dicom.boct_meta import boct_dicom_metadata
from oct_converter.dicom.e2e_meta import e2e_dicom_metadata
from oct_converter.dicom.fda_meta import fda_dicom_metadata
from oct_converter.dicom.fds_meta import fds_dicom_metadata
from oct_converter.dicom.img_meta import img_dicom_metadata
from oct_converter.dicom.metadata import DicomMetadata
from oct_converter.dicom.poct_meta import poct_dicom_metadata
from oct_converter.exceptions import InvalidOCTReaderError
from oct_converter.readers import BOCT, E2E, FDA, FDS, IMG, POCT

# Deterministic implementation UID based on package name and version
version = metadata.version("oct_converter")
implementation_uid = generate_uid(entropy_srcs=["oct_converter", version])


def opt_base_dicom(filepath: Path) -> Dataset:
    """Creates the base dicom to be populated.

    Args:
            filepath: Path to where output file is to be saved
    Returns:
            ds: FileDataset with file meta, preamble, and empty dataset
    """
    # Populate required values for file meta information
    file_meta = FileMetaDataset()
    file_meta.MediaStorageSOPClassUID = OphthalmicTomographyImageStorage
    file_meta.MediaStorageSOPInstanceUID = generate_uid()
    file_meta.ImplementationClassUID = implementation_uid
    file_meta.TransferSyntaxUID = ExplicitVRLittleEndian

    # Create the FileDataset instance with file meta, preamble and empty DS
    ds = FileDataset(str(filepath), {}, file_meta=file_meta, preamble=b"\0" * 128)
    ds.is_little_endian = True
    ds.is_implicit_VR = False  # Explicit VR
    return ds


def populate_patient_info(ds: Dataset, meta: DicomMetadata) -> Dataset:
    """Populates Patient Module PS3.3 C.7.1.1

    Args:
            ds: current dataset
            meta: DICOM metadata information
    Returns:
            ds: Dataset, updated with patient information
    """
    # Patient Module PS3.3 C.7.1.1
    ds.PatientName = f"{meta.patient_info.last_name}^{meta.patient_info.first_name}"
    ds.PatientID = meta.patient_info.patient_id
    ds.PatientSex = meta.patient_info.patient_sex
    ds.PatientBirthDate = (
        meta.patient_info.patient_dob.strftime("%Y%m%d")
        if meta.patient_info.patient_dob
        else ""
    )
    return ds


def populate_manufacturer_info(ds: Dataset, meta: DicomMetadata) -> Dataset:
    """Populates equipment modules PS3.3 C.7.5.1, PS3.3 C.7.5.2

    Args:
            ds: current dataset
            meta: DICOM metadata information
    Returns:
            ds: Dataset, updated with equipment information
    """
    # General and enhanced equipment module PS3.3 C.7.5.1, PS3.3 C.7.5.2
    ds.Manufacturer = meta.manufacturer_info.manufacturer
    ds.ManufacturerModelName = meta.manufacturer_info.manufacturer_model
    ds.DeviceSerialNumber = meta.manufacturer_info.device_serial
    ds.SoftwareVersions = meta.manufacturer_info.software_version

    # OPT parameter module PS3.3 C.8.17.9
    cd, cv, cm = meta.oct_image_params.opt_acquisition_device.value
    ds.AcquisitionDeviceTypeCodeSequence = [Dataset()]
    ds.AcquisitionDeviceTypeCodeSequence[0].CodeValue = cv
    ds.AcquisitionDeviceTypeCodeSequence[0].CodingSchemeDesignator = cd
    ds.AcquisitionDeviceTypeCodeSequence[0].CodeMeaning = cm
    ds.DetectorType = meta.oct_image_params.DetectorType.value
    return ds


def populate_opt_series(ds: Dataset, meta: DicomMetadata) -> Dataset:
    """Populates study and series modules, PS3.3 C.7.2.1, PS3.3 C.7.3.1,
    PS3.3 C.8.17.6, and PS3.3 C.12.1

    Args:
            ds: current dataset
            meta: DICOM metadata information
    Returns:
            ds: Dataset, updated with study and series information
    """
    # General study module PS3.3 C.7.2.1
    # Deterministic StudyInstanceUID based on study ID
    # ds.StudyInstanceUID = generate_uid(entropy_srcs=[
    # 	# str(uuid.uuid4()),
    # 	str(meta.series_info.study_id)
    # ])

    # # General series module PS3.3 C.7.3.1
    # ds.SeriesInstanceUID = generate_uid(entropy_srcs=[
    # 	# str(uuid.uuid4()),
    # 	str(meta.series_info.series_id)
    # ])
    ds.StudyInstanceUID = generate_uid()
    ds.SeriesInstanceUID = generate_uid()
    ds.Laterality = meta.series_info.laterality
    # Ophthalmic Tomography Series PS3.3 C.8.17.6
    ds.Modality = "OPT"
    ds.SeriesNumber = int(meta.series_info.series_id)

    # SOP Common module PS3.3 C.12.1
    ds.SOPClassUID = OphthalmicTomographyImageStorage
    ds.SOPInstanceUID = generate_uid()
    return ds


def populate_ocular_region(ds: Dataset, meta: DicomMetadata) -> Dataset:
    """Populates ocular region modules, PS3.3 C.8.17.5, PS3.3 C.7.6.16.2.8,
    and PS3.3 C.7.6.16.2.1

    Args:
            ds: current dataset
            meta: DICOM metadata information
    Returns:
            ds: Dataset, updated with ocular region information
    """
    # Ocular region imaged module PS3.3 C.8.17.5
    cd, cv, cm = meta.series_info.opt_anatomy.value
    ds.ImageLaterality = meta.series_info.laterality
    ds.AnatomicRegionSequence = [Dataset()]
    ds.AnatomicRegionSequence[0].CodeValue = cv
    ds.AnatomicRegionSequence[0].CodingSchemeDesignator = cd
    ds.AnatomicRegionSequence[0].CodeMeaning = cm
    return ds


def opt_shared_functional_groups(ds: Dataset, meta: DicomMetadata) -> Dataset:
    # ---- Shared
    shared_ds = [Dataset()]
    # Frame anatomy PS3.3 C.7.6.16.2.8
    shared_ds[0].FrameAnatomySequence = [Dataset()]
    shared_ds[0].FrameAnatomySequence[0] = ds.AnatomicRegionSequence[0].copy()
    shared_ds[0].FrameAnatomySequence[0].FrameLaterality = meta.series_info.laterality
    # Pixel Measures PS3.3 C.7.6.16.2.1
    shared_ds[0].PixelMeasuresSequence = [Dataset()]
    shared_ds[0].PixelMeasuresSequence[
        0
    ].PixelSpacing = meta.image_geometry.pixel_spacing
    shared_ds[0].PixelMeasuresSequence[
        0
    ].SliceThickness = meta.image_geometry.slice_thickness
    # Plane Orientation PS3.3 C.7.6.16.2.4
    shared_ds[0].PlaneOrientationSequence = [Dataset()]
    shared_ds[0].PlaneOrientationSequence[
        0
    ].ImageOrientationPatient = meta.image_geometry.image_orientation
    ds.SharedFunctionalGroupsSequence = shared_ds
    return ds


def write_opt_dicom(
    meta: DicomMetadata, frames: t.List[np.ndarray], filepath: Path
) -> Path:
    """Writes required DICOM metadata and oct pixel data to .dcm file.

    Args:
            meta: DICOM metadata information
            frames: list of frames of pixel data
            filepath: Path to where output file is being saved
    Returns:
            Path to created DICOM file
    """
    ds = opt_base_dicom(filepath)
    ds = populate_patient_info(ds, meta)
    ds = populate_manufacturer_info(ds, meta)
    ds = populate_opt_series(ds, meta)
    ds = populate_ocular_region(ds, meta)
    ds = opt_shared_functional_groups(ds, meta)

    # TODO: Frame of reference if fundus image present

    # OPT Image Module PS3.3 C.8.17.7
    ds.ImageType = ["DERIVED", "SECONDARY"]
    ds.SamplesPerPixel = 1
    ds.AcquisitionDateTime = (
        meta.series_info.acquisition_date.strftime("%Y%m%d%H%M%S.%f")
        if meta.series_info.acquisition_date
        else ""
    )
    ds.AcquisitionNumber = 1
    ds.PhotometricInterpretation = "MONOCHROME2"
    # Unsigned integer
    ds.PixelRepresentation = 0
    # Use 16 bit pixel
    ds.BitsAllocated = 16
    ds.BitsStored = ds.BitsAllocated
    ds.HighBit = ds.BitsAllocated - 1
    ds.SamplesPerPixel = 1
    ds.NumberOfFrames = len(frames)

    # Multi-frame Functional Groups Module PS3.3 C.7.6.16
    dt = datetime.now()
    ds.ContentDate = dt.strftime("%Y%m%d")
    timeStr = dt.strftime("%H%M%S.%f")  # long format with micro seconds
    ds.ContentTime = timeStr
    ds.InstanceNumber = 1

    per_frame = []
    pixel_data_bytes = list()
    # Normalize
    frames = normalize_volume(frames)
    # Convert to a 3d volume
    pixel_data = np.array(frames).astype(np.uint16)
    ds.Rows = pixel_data.shape[1]
    ds.Columns = pixel_data.shape[2]
    for i in range(pixel_data.shape[0]):
        # Per Frame Functional Groups
        frame_fgs = Dataset()
        frame_fgs.PlanePositionSequence = [Dataset()]
        ipp = [0, 0, i * meta.image_geometry.slice_thickness]
        frame_fgs.PlanePositionSequence[0].ImagePositionPatient = ipp
        frame_fgs.FrameContentSequence = [Dataset()]
        frame_fgs.FrameContentSequence[0].InStackPositionNumber = i + 1
        frame_fgs.FrameContentSequence[0].StackID = "1"

        # Pixel data
        frame_dat = pixel_data[i, :, :]
        pixel_data_bytes.append(frame_dat.tobytes())
        per_frame.append(frame_fgs)
    ds.PerFrameFunctionalGroupsSequence = per_frame
    ds.PixelData = pixel_data.tobytes()
    ds.save_as(filepath)
    return filepath


def write_fundus_dicom(
    meta: DicomMetadata, frames: t.List[np.ndarray], filepath: Path
) -> Path:
    """Writes required DICOM metadata and fundus pixel data to .dcm file.

    Args:
            meta: DICOM metadata information
            frames: list of frames of pixel data
            filepath: Path to where output file is being saved
    Returns:
            Path to created DICOM file
    """
    ds = opt_base_dicom(filepath)
    ds = populate_patient_info(ds, meta)
    ds = populate_manufacturer_info(ds, meta)
    ds = populate_opt_series(ds, meta)
    ds.Modality = "OP"
    ds = populate_ocular_region(ds, meta)
    ds = opt_shared_functional_groups(ds, meta)

    # OPT Image Module PS3.3 C.8.17.7
    ds.ImageType = ["DERIVED", "SECONDARY"]
    ds.SamplesPerPixel = 1
    ds.AcquisitionDateTime = (
        meta.series_info.acquisition_date.strftime("%Y%m%d%H%M%S.%f")
        if meta.series_info.acquisition_date
        else ""
    )
    ds.AcquisitionNumber = 1
    ds.PhotometricInterpretation = "MONOCHROME2"
    # Unsigned integer
    ds.PixelRepresentation = 0
    # Use 16 bit pixel
    ds.BitsAllocated = 16
    ds.BitsStored = ds.BitsAllocated
    ds.HighBit = ds.BitsAllocated - 1
    ds.SamplesPerPixel = 1
    ds.NumberOfFrames = 1

    # Multi-frame Functional Groups Module PS3.3 C.7.6.16
    dt = datetime.now()
    ds.ContentDate = dt.strftime("%Y%m%d")
    timeStr = dt.strftime("%H%M%S.%f")  # long format with micro seconds
    ds.ContentTime = timeStr
    ds.InstanceNumber = 1
    pixel_data = np.array(frames).astype(np.uint16)
    ds.Rows = pixel_data.shape[0]
    ds.Columns = pixel_data.shape[1]

    ds.PixelData = pixel_data.tobytes()
    ds.save_as(filepath)
    return filepath


def write_color_fundus_dicom(
    meta: DicomMetadata, frames: t.List[np.ndarray], filepath: Path
) -> Path:
    """Writes required DICOM metadata and RGB fundus pixel data to .dcm file.

    Args:
            meta: DICOM metadata information
            frames: list of frames of pixel data
            filepath: Path to where output file is being saved
    Returns:
            Path to created DICOM file
    """
    ds = opt_base_dicom(filepath)
    ds = populate_patient_info(ds, meta)
    ds = populate_manufacturer_info(ds, meta)
    ds = populate_opt_series(ds, meta)
    ds.Modality = "OP"
    ds = populate_ocular_region(ds, meta)
    ds = opt_shared_functional_groups(ds, meta)

    # OPT Image Module PS3.3 C.8.17.7
    ds.ImageType = ["DERIVED", "SECONDARY"]
    ds.SamplesPerPixel = 1
    ds.AcquisitionDateTime = (
        meta.series_info.acquisition_date.strftime("%Y%m%d%H%M%S.%f")
        if meta.series_info.acquisition_date
        else ""
    )
    ds.AcquisitionNumber = 1
    ds.PhotometricInterpretation = "RGB"
    # Unsigned integer
    ds.PixelRepresentation = 0
    # Use 16 bit pixel
    ds.BitsAllocated = 16
    ds.BitsStored = ds.BitsAllocated
    ds.HighBit = ds.BitsAllocated - 1
    ds.SamplesPerPixel = 1
    ds.NumberOfFrames = 1

    # Multi-frame Functional Groups Module PS3.3 C.7.6.16
    dt = datetime.now()
    ds.ContentDate = dt.strftime("%Y%m%d")
    timeStr = dt.strftime("%H%M%S.%f")  # long format with micro seconds
    ds.ContentTime = timeStr
    ds.InstanceNumber = 1

    pixel_data = np.array(frames).astype(np.uint16)
    ds.Rows = pixel_data.shape[0]
    ds.Columns = pixel_data.shape[1]

    ds.PixelData = pixel_data.tobytes()
    ds.save_as(filepath)
    return filepath


def create_dicom_from_oct(
    input_file: str,
    output_dir: str = None,
    rows: int = 1024,
    cols: int = 512,
    interlaced: bool = False,
    diskbuffered: bool = False,
) -> list:
    """Creates a DICOM file with the data parsed from
    the input file.

    Args:
            input_file: File with OCT data, .fda/.fds/.img/.e2e/.OCT
            output_dir: Output directory, will be created if
            not currently exists. Default None places file in
            current working directory.
            rows: If .img file, allows for manually setting rows
            cols: If .img file, allows for manually setting cols
            interlaced: If .img file, allows for setting interlaced
            diskbuffered: If Bioptigen .OCT, allows for setting diskbuffered

    Returns:
            list: list of Path(s) to DICOM file
    """
    if output_dir:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
    else:
        output_dir = Path.cwd()

    file_suffix = input_file.split(".")[-1].lower()

    if file_suffix == "fds":
        files = create_dicom_from_fds(input_file, output_dir)
    elif file_suffix == "fda":
        files = create_dicom_from_fda(input_file, output_dir)
    elif file_suffix == "img":
        files = create_dicom_from_img(input_file, output_dir, rows, cols, interlaced)
    elif file_suffix == "oct":
        # Bioptigen and Octovue both use .OCT.
        # BOCT._validate on init can check if Bioptigen, else Optivue
        try:
            BOCT(input_file)
            files = create_dicom_from_boct(input_file, output_dir, diskbuffered)
        except (InvalidOCTReaderError, StreamError):
            # if BOCT raises, treat as POCT
            files = create_dicom_from_poct(input_file, output_dir)
    elif file_suffix == "e2e":
        files = create_dicom_from_e2e(input_file, output_dir)
    else:
        raise TypeError(
            f"DICOM conversion for {file_suffix} is not supported. "
            "Currently supported filetypes are .e2e, .fds, .fda, .img, .OCT."
        )

    return files


def normalize_volume(vol: list[np.ndarray]) -> list[np.ndarray]:
    """Normalizes pixel intensities within a range of 0-100.

    Args:
        vol: List of frames
    Returns:
        Normalized list of frames
    """
    arr = np.array(vol)
    norm_vol = []
    diff_arr = arr.max() - arr.min()
    for i in arr:
        temp = ((i - arr.min()) / diff_arr) * 100
        norm_vol.append(temp)
    return norm_vol


def create_dicom_from_boct(
    input_file: str,
    output_dir: str = None,
    diskbuffered: bool = False,
) -> list:
    """Creates DICOM file(s) with the data parsed from
    the input file.

    Args:
            input_file: Bioptigen OCT file
            output_dir: Output directory
            diskbuffered: If True, reduces memory usage by storing volume on disk using HDF5.

    Returns:
            list: List of path(s) to DICOM file(s)"""

    boct = BOCT(input_file)
    oct_volumes = boct.read_oct_volume(diskbuffered)
    if len(oct_volumes) == 0:
        raise ValueError("No OCT volumes found in OCT input file.")

    files = []

    for count, oct in enumerate(oct_volumes):
        meta = boct_dicom_metadata(oct)
        filename = f"{Path(input_file).stem}_{str(count)}.dcm"
        filepath = Path(output_dir, filename)
        file = write_opt_dicom(meta, oct.volume, filepath)
        files.append(file)

    return files


def create_dicom_from_e2e(
    input_file: str,
    output_dir: str = None,
) -> list:
    """Creates DICOM file(s) with the data parsed from
    the input file.

    Args:
            input_file: E2E file with OCT data
            output_dir: Output directory

    Returns:
            list: List of path(s) to DICOM file(s)
    """
    e2e = E2E(input_file)
    oct_volumes = e2e.read_oct_volume()
    fundus_images = e2e.read_fundus_image()
    if len(oct_volumes) == 0 and len(fundus_images) == 0:
        raise ValueError("No OCT volumes or fundus images found in e2e input file.")

    files = []

    if len(fundus_images) > 0:
        for count, fundus in enumerate(fundus_images):
            meta = e2e_dicom_metadata(fundus)
            filename = f"{Path(input_file).stem}_fundus_{str(count)}.dcm"
            filepath = Path(output_dir, filename)
            file = write_fundus_dicom(meta, fundus.image, filepath)
            files.append(file)

    if len(oct_volumes) > 0:
        for count, oct in enumerate(oct_volumes):
            meta = e2e_dicom_metadata(oct)
            filename = f"{Path(input_file).stem}_oct_{str(count)}.dcm"
            filepath = Path(output_dir, filename)
            file = write_opt_dicom(meta, oct.volume, filepath)
            files.append(file)

    return files


def create_dicom_from_fda(
    input_file: str,
    output_dir: str,
) -> list:
    """Creates DICOM file(s) with the data parsed from
    the input file.

    Args:
            input_file: FDA file with OCT data
            output_dir: Output directory

    Returns:
            list: List of path(s) to DICOM file(s)
    """
    files = []
    fda = FDA(input_file)
    oct = fda.read_oct_volume()
    meta = fda_dicom_metadata(oct)
    output_filename = f"{Path(input_file).stem}.dcm"
    filepath = Path(output_dir, output_filename)
    file = write_opt_dicom(meta, oct.volume, filepath)
    files.append(file)

    # Attempt to parse fundus images
    fundus = fda.read_fundus_image()
    if fundus:
        output_filename = f"{Path(input_file).stem}_fundus.dcm"
        filepath = Path(output_dir, output_filename)
        meta.image_geometry.pixel_spacing = [1, 1]
        file = write_color_fundus_dicom(meta, fundus.image, filepath)
        files.append(file)

    fundus_grayscale = fda.read_fundus_image_gray_scale()
    if fundus_grayscale:
        output_filename = f"{Path(input_file).stem}_fundus_grayscale.dcm"
        filepath = Path(output_dir, output_filename)
        meta.image_geometry.pixel_spacing = [1, 1]
        file = write_fundus_dicom(meta, fundus_grayscale.image, filepath)
        files.append(file)

    return files


def create_dicom_from_fds(
    input_file: str,
    output_dir: str,
) -> list:
    """Creates DICOM file(s) with the data parsed from
    the input file.

    Args:
            input_file: FDS file with OCT data
            output_dir: Output directory

    Returns:
            list: List of path(s) to DICOM file(s)
    """
    files = []
    fds = FDS(input_file)
    oct = fds.read_oct_volume()
    meta = fds_dicom_metadata(oct)
    output_filename = f"{Path(input_file).stem}.dcm"
    filepath = Path(output_dir, output_filename)
    file = write_opt_dicom(meta, oct.volume, filepath)
    files.append(file)

    # Attempt to parse fundus images
    fundus = fds.read_fundus_image()
    if fundus:
        output_filename = f"{Path(input_file).stem}_fundus.dcm"
        filepath = Path(output_dir, output_filename)
        file = write_color_fundus_dicom(meta, fundus.image, filepath)
        files.append(file)

    return files


def create_dicom_from_img(
    input_file: str,
    output_dir: str,
    rows: int = 1024,
    cols: int = 512,
    interlaced: bool = False,
) -> Path:
    """Creates a DICOM file with the data parsed from
    the input file.

    Args:
            input_file: .img file with OCT data
            output_dir: Output directory
            rows: Optional, for manually setting rows. Default 1024.
            cols: Optional, for manually setting cols. Default 512.
            interlaced: Optional, for setting interlaced. Default False.

    Returns:
            list: List of path(s) to DICOM file(s)
    """
    img = IMG(input_file)
    oct = img.read_oct_volume(rows, cols, interlaced)
    meta = img_dicom_metadata(oct)
    output_filename = f"{Path(input_file).stem}.dcm"
    filepath = Path(output_dir, output_filename)
    file = write_opt_dicom(meta, oct.volume, filepath)
    return [file]


def create_dicom_from_poct(
    input_file: str,
    output_dir: str,
) -> list:
    """Creates DICOM file(s) with the data parsed from
    the input file.

    Args:
            input_file: File with POCT data
            output_dir: Output directory

    Returns:
            list: List of path(s) to DICOM file(s)
    """
    poct = POCT(input_file)
    octs = poct.read_oct_volume()
    files = []
    for count, oct in enumerate(octs):
        meta = poct_dicom_metadata(oct)
        filename = f"{Path(input_file).stem}_{str(count)}.dcm"
        filepath = Path(output_dir, filename)
        file = write_opt_dicom(meta, oct.volume, filepath)
        files.append(file)

    return files



from oct_converter.dicom import create_dicom_from_oct
from oct_converter.readers import BOCT

filepath = "../sample_files/sample.OCT"
boct = BOCT(filepath)
oct_volumes = boct.read_oct_volume(
    diskbuffered=True
)  # returns an OCT volume with additional metadata if available
for oct in oct_volumes:
    oct.peek()  # plots a montage of the volume
    oct.save("boct_testing.avi")  # save volume as a movie
    oct.save(
        "boct_testing.png"
    )  # save volume as a set of sequential images, fds_testing_[1...N].png

# create DICOM from .OCT
dcm = create_dicom_from_oct(filepath)
# Output dir can be specified, otherwise will
# default to current working directory.
# If multiple volumes are identified within the file,
# multiple DICOMs will be outputted.
# Additionally, diskbuffered can be specified to store
# volume on disk using HDF5 to reduce memory usage
dcm = create_dicom_from_oct(filepath, diskbuffered=True)



from oct_converter.readers import Dicom

filepath = "../sample_files/sample_dcm.dcm"
file = Dicom(filepath)
oct_volume = (
    file.read_oct_volume()
)  # returns an OCT volume with additional metadata if available

oct_volume.save("dcm_testing.avi")



"""Init module."""

from .fundus import FundusImageWithMetaData
from .oct import OCTVolumeWithMetaData

__all__ = [
    "version",
    "implementation_uid",
    "FundusImageWithMetaData",
    "OCTVolumeWithMetaData",
]



from __future__ import annotations

from pathlib import Path

from oct_converter.image_types import OCTVolumeWithMetaData


class Dicom(object):
    def __init__(self, filepath: str | Path) -> None:
        self.filepath = Path(filepath)
        if not self.filepath.exists():
            raise FileNotFoundError(self.filepath)

    def read_oct_volume(self) -> OCTVolumeWithMetaData:
        """Reads OCT data.

        Returns:
            OCTVolumeWithMetaData
        """
        import pydicom

        dicom_data = pydicom.dcmread(self.filepath)
        if dicom_data.Manufacturer.startswith("Carl Zeiss Meditec"):
            raise ValueError(
                "This appears to be a Zeiss DCM. You may need to read with the ZEISSDCM class."
            )
        pixel_data = dicom_data.pixel_array
        oct_volume = OCTVolumeWithMetaData(volume=pixel_data)
        return oct_volume
